// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// InfoMP3.cpp
//
// $Id: InfoMP3.cpp,v 1.5 2002/02/26 19:30:26 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "InfoMP3.h"

#include <id3/misc_support.h>
#include <xb/ErrorManager.h>
#include <xb/FileUtil.h>
#include <xb/String.h>

#include "MP3Info.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

InfoMP3::InfoMP3() :
	Info(),
	m_pTag(NULL)
{

}

//------------------------------------------------------------------------------

InfoMP3::~InfoMP3()
{
	delete m_pTag;
}

//------------------------------------------------------------------------------

bool
InfoMP3::CanReadFile(const xb::Filename& strFilename) const
{
	return xb::FileUtil::MatchesExtension(strFilename, ".mp3");
}

//------------------------------------------------------------------------------

void
InfoMP3::Init(const xb::Filename& strFilename)
{
	Assert(m_pTag == NULL, "m_pTag not null");
	
	Info::Init(strFilename);
	m_pTag = new ID3_Tag(strFilename.c_str());
}

//------------------------------------------------------------------------------

int
InfoMP3::GetSeconds() const
{
	long nResult = GetTagInt(ID3FID_SONGLEN) / 1000;
	if (nResult == 0) {
		nResult = CalculateTime();
	}

	return nResult;
}

//------------------------------------------------------------------------------

void
InfoMP3::SetSeconds(int nSec)
{
	long nMilliseconds = 1000 * GetSeconds();
	xb::String strTime(nMilliseconds);
	SetTag(ID3FID_SONGLEN, strTime);
	m_bSaveNeeded = true;
}

//------------------------------------------------------------------------------

std::string
InfoMP3::GetTitle() const
{
	return GetTagString(ID3FID_TITLE);
}

//------------------------------------------------------------------------------

void
InfoMP3::SetTitle(const std::string& s)
{
	AssertNotNull(m_pTag);
	ID3_AddTitle(m_pTag, s.c_str(), true);
	m_bSaveNeeded = true;
}

//------------------------------------------------------------------------------

std::string
InfoMP3::GetArtist() const
{
	return GetTagString(ID3FID_LEADARTIST);
}

//------------------------------------------------------------------------------

void
InfoMP3::SetArtist(const std::string& s)
{
	AssertNotNull(m_pTag);
	ID3_AddArtist(m_pTag, s.c_str(), true);
	m_bSaveNeeded = true;
}

//------------------------------------------------------------------------------

std::string
InfoMP3::GetAlbum() const
{
	return GetTagString(ID3FID_ALBUM);
}

//------------------------------------------------------------------------------

void
InfoMP3::SetAlbum(const std::string& s)
{
	AssertNotNull(m_pTag);
	ID3_AddAlbum(m_pTag, s.c_str(), true);
	m_bSaveNeeded = true;
}

//------------------------------------------------------------------------------

std::string
InfoMP3::GetTrack() const
{
	return GetTagString(ID3FID_TRACKNUM);
}

//------------------------------------------------------------------------------

void
InfoMP3::SetTrack(const std::string& s)
{
	AssertNotNull(m_pTag);
	ID3_AddTrack(m_pTag, atoi(s.c_str()), 0, true);
	m_bSaveNeeded = true;
}
  
//------------------------------------------------------------------------------

void
InfoMP3::Dump() const
{
	EM::I()->Trace() << "InfoMP3 -- Dump:" << endl;
	Info::Dump();
}

//------------------------------------------------------------------------------

ID3_Field*
InfoMP3::GetFieldPtr(ID3_FrameID eFrameID) const
{
	ID3_Field* pResult = NULL;

	ID3_Frame* pFrame = m_pTag->Find(eFrameID);
	if (pFrame != NULL) {

		pResult = pFrame->GetField(ID3FN_TEXT);

	}
	// else {
	// 	EM::I()->Error("No frame found for " + xb::String((int)eFrameID));
	//	EM::I()->Error("Avaiable frames:");
	//	Dump();
	// }

	return pResult;
}

//------------------------------------------------------------------------------

std::string
InfoMP3::GetTagString(ID3_FrameID eFrameID) const
{
	std::string strResult;
	
	ID3_Field* pField = GetFieldPtr(eFrameID);
	if (pField != NULL) {
		strResult = pField->GetRawText();
	}
	
	return strResult;
}

//------------------------------------------------------------------------------

int
InfoMP3::GetTagInt(ID3_FrameID eFrameID) const
{
	int nResult = 0;
	
	std::string s = GetTagString(eFrameID);
	if (s.length() > 0) {
		nResult = atoi(s.c_str());
	}

	return nResult;
}

//------------------------------------------------------------------------------

long
InfoMP3::CalculateTime() const
{
	MP3Info info(m_strFilename);
	return info.GetSeconds();
}

//------------------------------------------------------------------------------

void
InfoMP3::Save()
{
	AssertNotNull(m_pTag);
	m_pTag->Strip(ID3TT_ID3V1);
	m_pTag->Update(ID3TT_ID3V2);
	m_bSaveNeeded = false;
}

//------------------------------------------------------------------------------

void
InfoMP3::SetTag(ID3_FrameID eFrameID, const std::string& strText)
{
	ID3_Field* pField = GetFieldPtr(eFrameID);
	if (pField != NULL) {
		pField->Set(strText.c_str());
	}
	else {
		ID3_Frame* pFrame = new ID3_Frame(eFrameID);
		AssertNotNull(pFrame);

		pFrame->GetField(ID3FN_TEXT)->Set(strText.c_str());
		AssertNotNull(m_pTag);
		m_pTag->AttachFrame(pFrame);

		// PATRICK:
		// delete pFrame ???
	}
}

//------------------------------------------------------------------------------

extern "C" {
	Info* maker()
	{
		return new InfoMP3;
	}
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
