// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// InfoOgg.cpp
//
// $Id: InfoOgg.cpp,v 1.4 2002/02/26 19:30:26 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "InfoOgg.h"

#include <vorbis/vorbisfile.h>
#include <xb/ErrorManager.h>
#include <xb/FileUtil.h>
#include <xb/String.h>

#include "PropList.h"
#include "System.h"
#include "vcedit.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

InfoOgg::InfoOgg() :
	Info(),
	m_bInitOK(false),
	m_nSeconds(0)
{

}

//------------------------------------------------------------------------------

InfoOgg::~InfoOgg()
{
}

//------------------------------------------------------------------------------


bool
InfoOgg::CanReadFile(const xb::Filename& strFilename) const
{
	return xb::FileUtil::MatchesExtension(strFilename, ".ogg");
}

//------------------------------------------------------------------------------

void
InfoOgg::Init(const xb::Filename& strFilename)
{
	Info::Init(strFilename);

	m_bInitOK = false;
	
	FILE* pStream = fopen(m_strFilename.c_str(), "r");
	if (pStream == NULL) {
		EM::I()->Warn() << "couldn't open " << m_strFilename << endl;
		return;
	}

	OggVorbis_File vf;
	if (ov_open(pStream, &vf, NULL, 0) < 0) {
		EM::I()->Warn() << "InfoOgg::Init() -- ov_open failed." << endl;
		fclose(pStream);
	}
	else {
		m_nSeconds = (int)ov_time_total(&vf, -1);

		EM::I()->Trace() << "seconds = " << m_nSeconds << endl;
		
		vorbis_comment* pComment = ov_comment(&vf, -1);

		EM::I()->Trace() << "got comment, querying:" << endl;
		
		if (pComment != NULL) {
			m_strTitle = GetComment(pComment, "title");
			m_strArtist = GetComment(pComment, "artist");
			m_strAlbum = GetComment(pComment, "album");
			m_strTrack = GetComment(pComment, "tracknumber");
		}

		EM::I()->Trace() << "clearing vf..." << endl;

		ov_clear(&vf);
	}

	EM::I()->Trace() << "Init done." << endl;
}

//------------------------------------------------------------------------------

int
InfoOgg::GetSeconds() const
{
	return m_nSeconds;
}

//------------------------------------------------------------------------------

void
InfoOgg::SetSeconds(int nSec)
{
	EM::I()->Warn() << "can't set seconds on an ogg" << endl;
}

//------------------------------------------------------------------------------

std::string
InfoOgg::GetTitle() const
{
	return m_strTitle;
}

//------------------------------------------------------------------------------

void
InfoOgg::SetTitle(const std::string& s)
{
	m_strTitle = s;
	m_bSaveNeeded = true;
}

//------------------------------------------------------------------------------

std::string
InfoOgg::GetArtist() const
{
	return m_strArtist;
}

//------------------------------------------------------------------------------

void
InfoOgg::SetArtist(const std::string& s)
{
	m_strArtist = s;
	m_bSaveNeeded = true;
}

//------------------------------------------------------------------------------

std::string
InfoOgg::GetAlbum() const
{
	return m_strAlbum;
}

//------------------------------------------------------------------------------

void
InfoOgg::SetAlbum(const std::string& s)
{
	m_strAlbum = s;
	m_bSaveNeeded = true;
}

//------------------------------------------------------------------------------

std::string
InfoOgg::GetTrack() const
{
	return m_strTrack;
}

//------------------------------------------------------------------------------

void
InfoOgg::SetTrack(const std::string& s)
{
	m_strTrack = s;
	m_bSaveNeeded = true;
}
  
//------------------------------------------------------------------------------

void
InfoOgg::Save()
{
	vcedit_state* pState = vcedit_new_state();

	FILE* pStream = fopen(m_strFilename.c_str(), "rb");
	if (pStream == NULL) {
		EM::I()->Error() << "couldn't open " << m_strFilename << endl;
		return;
	}

	if (vcedit_open(pState, pStream) < 0) {
		EM::I()->Error() << "failed vorbis open: " << vcedit_error(pState) <<
			endl;
		EM::I()->Error() << "filename = " << m_strFilename << endl;
		return;
	}

	// get and clear existing comments
	vorbis_comment* pVC = vcedit_comments(pState);
	vorbis_comment_clear(pVC);
	vorbis_comment_init(pVC);

	// set comments
	vorbis_comment_add_tag(pVC, "artist", (char *)m_strArtist.c_str());
	vorbis_comment_add_tag(pVC, "title", (char *)m_strTitle.c_str());
	vorbis_comment_add_tag(pVC, "album", (char *)m_strAlbum.c_str());
	vorbis_comment_add_tag(pVC, "tracknumber", (char *)m_strTrack.c_str());

	// open out file
	xb::Filename strOutFile = System::GetPropList()->GetValue("SpoolRoot") + "tmp.ogg";
	EM::I()->Trace() << "opening " << strOutFile << " for temp ogg output..." <<
		endl;
	FILE* pStreamOut = fopen(strOutFile.c_str(), "wb");
	if (pStreamOut == NULL) {
		EM::I()->Error() << "couldn't open " << strOutFile << endl;
		return;
	}

	// write out entire stream (including new comments)
	if (vcedit_write(pState, pStreamOut) < 0) {
		EM::I()->Error() << "failed vorbis comment write: " <<
			vcedit_error(pState) << endl;
		return;
	}

	vcedit_clear(pState);

	fclose(pStream);
	fclose(pStreamOut);

	// move tmp file to actual file...
	xb::FileUtil::Rename(strOutFile, m_strFilename);

	EM::I()->Trace() << "InfoOgg::Save() -- success" << endl;
}

//------------------------------------------------------------------------------

void
InfoOgg::Dump() const
{
	EM::I()->Trace() << "InfoOgg -- Dump:" << endl;
	Info::Dump();
}

//------------------------------------------------------------------------------

std::string
InfoOgg::GetComment(vorbis_comment* pComment, const std::string& strTag)
{
	std::string strResult;
	
	char* s = vorbis_comment_query(pComment, (char *)strTag.c_str(), 0);
	if (s != NULL) {
		strResult = s;
	}

	return strResult;
}

//------------------------------------------------------------------------------

extern "C" {
	Info* maker()
	{
		return new InfoOgg;
	}
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
