// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// PlayerMPG.cpp
//
// $Id: PlayerMPG.cpp,v 1.18 2002/02/26 19:30:26 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "PlayerMPG.h"

#include <math.h>
#include <xb/ErrorManager.h>

#include "ID3.h"
#include "Layer.h"
#include "Layer1.h"
#include "Layer2.h"
#include "Layer3.h"
#include "Output.h"
#include "PlayerInfo.h"
#include "Synth.h"
#include "Synth1to1.h"
#include "Visualization.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

const int PlayerMPG::SBLIMIT;

//------------------------------------------------------------------------------

PlayerMPG::PlayerMPG() :
	Player(),
	m_pInfo(NULL),
	m_bHaveXingHeader(false),
	m_pPCMSample(NULL),
	m_pLayer(NULL)
{
}

//------------------------------------------------------------------------------

PlayerMPG::~PlayerMPG()
{
	delete m_pInfo;
	delete m_pLayer;
	delete m_pPCMSample;
}

//------------------------------------------------------------------------------

void
PlayerMPG::Init(Output* pOutput, Visualization* pVis)
{
	Player::Init(pOutput, pVis);
	
	// m_pOutput = pOutput;
	// m_pVisualization = pVis;
	
	m_nResolution = 16;
	m_nChannels = 2;
	m_nDownsample = 0;
	m_nDownsampleCustom = 44100;
}

//------------------------------------------------------------------------------

void
PlayerMPG::Play(const xb::Filename& strFilename, int nTrack)
{
	EM::I()->Trace() << "PlayerMPG::Play -- file = " << strFilename << endl;

	m_nTrack = nTrack;
	
	m_Frame.Reset();
	m_FrameTemp.Reset();

	m_pInfo = new PlayerInfo;
	m_pInfo->going = 1;
	m_pInfo->first_frame = TRUE;
	m_pInfo->output_audio = TRUE;
	m_pInfo->jump_to_time = -1;
	m_pInfo->eof = 0;
	m_pInfo->resolution = m_nResolution;
	m_nSkipFrames = 0;
	m_bAudioError = false;
	m_bHaveXingHeader = false;

	m_strFilename = strFilename;

	m_bPlaying = true;
	
	Start();
}

//------------------------------------------------------------------------------

void
PlayerMPG::Stop(bool bSendSignal = true)
{
	if (m_pInfo && m_pInfo->going) {
		std::string s;
		m_strFilename = s;
		m_pInfo->going = FALSE;
		Terminate();
		delete m_pInfo;
		m_pInfo = NULL;
		m_pOutput->Close();
		delete m_pLayer;
		m_pLayer = NULL;

		delete m_pPCMSample;
		m_pPCMSample = NULL;
		
		m_bPlaying = false;
	}
}

//------------------------------------------------------------------------------

void
PlayerMPG::Pause(bool bPause)
{
	m_pOutput->Pause(bPause);
}

//------------------------------------------------------------------------------

void
PlayerMPG::Seek(int nTime)
{
	m_pInfo->jump_to_time = nTime;

	while (m_pInfo->jump_to_time != -1) {
		Thread::usleep(10000);
	}
}

//------------------------------------------------------------------------------

void
PlayerMPG::SeekPercentage(float fPercent)
{
	AssertNotNull(m_pInfo);

	int nTime = (int)(m_nLength * fPercent / 1000.0);
	fprintf(stderr, "m_nLength = %d, fPercent = %f, nTime = %d\n",
			m_nLength, fPercent, nTime);
	Seek(nTime);
}

//------------------------------------------------------------------------------

int
PlayerMPG::GetTime() const
{
	if (m_bAudioError) {
		return -2;
	}
	if (m_pInfo == NULL) {
		EM::I()->Trace() << "PlayerMPG::GetTime() -- m_pInfo == NULL" << endl;
		return -1;
	}
	if (!m_pInfo->going ||
		(m_pInfo->eof && !m_pOutput->BufferPlaying())) {
		EM::I()->Trace() << "PlayerMPG::GetTime() -- going == " <<
			m_pInfo->going << ", eof == " << m_pInfo->eof << endl;
		return -1;
	}

	return m_pOutput->ElapsedTime();
}

//------------------------------------------------------------------------------

bool
PlayerMPG::CanPlayFile(const xb::Filename& strFilename)
{
	bool bResult = false;
	
	char* ext = strrchr(strFilename.c_str(), '.');
	if (ext) {
		if (!strncasecmp(ext, ".mp3", 4)) {
			bResult = true;
		}
	}

	return bResult;
}

//------------------------------------------------------------------------------

bool
PlayerMPG::CheckCapability(std::string& strMessage)
{
	return true;
}

//------------------------------------------------------------------------------

void
PlayerMPG::Run()
{
	bool bOutputOpened = false;
	int nDispCount = 0;

	m_nBitRate = 0;

	if (m_pPCMSample != NULL) {
		delete m_pPCMSample;
	}
	m_pPCMSample = new unsigned char[32768];
	
	m_nPCMPoint = 0;

	ReadFrameInit();

	OpenStream();

	if (m_pInfo->eof || !ReadFrame(&m_Frame)) {
		m_pInfo->eof = TRUE;
	}

	if (!m_pInfo->eof) {

		if (m_nChannels == 2) {
			m_Frame.single = -1;
		}
		else {
			m_Frame.single = 3;
		}

		m_Frame.down_sample = m_nDownsample;
		m_Frame.down_sample_sblimit = SBLIMIT >> m_nDownsample;
		
		SetSynthFunctions(&m_Frame);

		if (m_pLayer != NULL) {
			m_pLayer->Init(m_Frame.down_sample_sblimit);
		}

		m_pInfo->tpf = m_Frame.ComputeTPF();

		for (;;) {

			// memcpy(&m_FrameTemp, &m_Frame, sizeof(frame));
			m_FrameTemp = m_Frame;

			if (!ReadFrame(&m_FrameTemp)) {
				m_pInfo->eof = TRUE;
				break;
			}

			if (m_Frame.lay != m_FrameTemp.lay ||
				m_Frame.sampling_frequency != m_FrameTemp.sampling_frequency ||
				m_Frame.stereo != m_FrameTemp.stereo ||
				m_Frame.lsf != m_FrameTemp.lsf) {

				m_Frame = m_FrameTemp;
			}
			else {
				break;
			}
		}

		m_pInfo->num_frames = CalcNumFrames(&m_Frame);

		m_Frame = m_FrameTemp;
		m_nBitRate = m_nDispBitRate =
			Frame::s_TabSel[m_Frame.lsf][m_Frame.lay - 1][m_Frame.bitrate_index];
		m_nFrequency = Frame::s_Freqs[m_Frame.sampling_frequency];
		m_nStereo = m_Frame.stereo;
		m_nLayer = m_Frame.lay;
		m_nLSF = m_Frame.lsf;
		m_bMpeg25 = m_Frame.mpeg25;
		m_nMode = m_Frame.mode;

		m_nLength = (int) (m_pInfo->num_frames * m_pInfo->tpf * 1000);

		Dump();
		
		/*
		if (!mpg123_title)
			mpg123_title = get_song_title(NULL,filename);
		
		mpg123_ip.set_info(mpg123_title, m_nLength, mpg123_bitrate * 1000, mpg123_freqs[m_Frame.sampling_frequency], m_Frame.stereo);
		*/

		bOutputOpened = true;
		if (!m_pOutput->Open(
			m_nResolution == 16 ? Output::FMT_S16_NE : Output::FMT_U8,
			Frame::s_Freqs[m_Frame.sampling_frequency] >> m_nDownsample,
			m_nChannels == 2 ? m_Frame.stereo : 1,
			m_nTrack))
		{
			m_bAudioError = TRUE;
			m_pInfo->eof = TRUE;
		}
		else {
			PlayFrame(&m_Frame);
		}
	}
	
	m_pInfo->first_frame = FALSE;
	while (m_pInfo->going)
	{
		if (m_pInfo->jump_to_time != -1)
		{
			if (!m_bHaveXingHeader)
				StreamJumpToFrame(
					&m_Frame, (int) (m_pInfo->jump_to_time / m_pInfo->tpf));
			else
				StreamJumpToByte(
					&m_Frame,
					SeekPoint(
						m_acXingTOC, xing_header.bytes,
						((double) m_pInfo->jump_to_time * 100.0) /
						((double) m_pInfo->num_frames * m_pInfo->tpf)));
			
			m_pOutput->Flush(m_pInfo->jump_to_time * 1000);
			m_pInfo->jump_to_time = -1;
			m_pInfo->eof = FALSE;

		}
		if (!m_pInfo->eof)
		{
			if (ReadFrame(&m_Frame))
			{
				if (m_Frame.lay != m_nLayer || m_Frame.lsf != m_nLSF)
				{
					m_FrameTemp = m_Frame;
					if (ReadFrame(&m_FrameTemp))
					{
						if (m_Frame.lay == m_FrameTemp.lay &&
							m_Frame.lsf == m_FrameTemp.lsf)
						{
							m_nLayer = m_Frame.lay;
							m_nLSF = m_Frame.lsf;
							m_Frame = m_FrameTemp;
							SetSynthFunctions(&m_Frame);
						}
						else
						{
							m_Frame = m_FrameTemp;
							m_nSkipFrames = 2;
							m_pInfo->output_audio = FALSE;
							continue;
						}
						
					}
				}
				if (Frame::s_Freqs[m_Frame.sampling_frequency] !=
					m_nFrequency || m_nStereo != m_Frame.stereo)
				{
					// memcpy(&m_FrameTemp,&m_Frame,sizeof(struct frame));
					m_FrameTemp = m_Frame;
					if (ReadFrame(&m_FrameTemp))
					{
						if (m_Frame.sampling_frequency == m_FrameTemp.sampling_frequency && m_FrameTemp.stereo == m_Frame.stereo)
						{
							m_pOutput->BufferFreeAmt();
							m_pOutput->BufferFreeAmt();
							
							while(m_pOutput->BufferPlaying() &&
								  m_pInfo->going &&
								  m_pInfo->jump_to_time == -1) {
								Thread::usleep(20000);
							}
							
							if(!m_pInfo->going)
								break;
							
							int nTempTime = m_pOutput->ElapsedTime();
							m_pOutput->Close();
							m_nFrequency = Frame::s_Freqs[m_Frame.sampling_frequency];
							m_nStereo = m_Frame.stereo;
							if (!m_pOutput->Open(
								m_nResolution == 16 ?
								Output::FMT_S16_NE : Output::FMT_U8,
								Frame::s_Freqs[m_Frame.sampling_frequency] >> m_nDownsample,
				  				m_nChannels == 2 ? m_Frame.stereo : 1,
								m_nTrack))
							{
								m_bAudioError = TRUE;
								m_pInfo->eof = TRUE;
							}
							m_pOutput->Flush(nTempTime);
							// PATRICK: fix
							// mpg123_ip.set_info(mpg123_title, m_nLength, mpg123_bitrate * 1000, m_nFrequency, m_nStereo);

							m_Frame = m_FrameTemp;
							SetSynthFunctions(&m_Frame);
						}
						else
						{
							m_Frame = m_FrameTemp;
							m_nSkipFrames = 2;
							m_pInfo->output_audio = FALSE;
							continue;
						}
					}					
				}
				
				if (Frame::s_TabSel[m_Frame.lsf][m_Frame.lay - 1][m_Frame.bitrate_index] != m_nBitRate)
					m_nBitRate = Frame::s_TabSel[m_Frame.lsf][m_Frame.lay - 1][m_Frame.bitrate_index];
				
				if (!nDispCount)
				{
					nDispCount = 20;
					if (m_nBitRate != m_nDispBitRate)
					{
						m_nDispBitRate = m_nBitRate;
						m_pInfo->num_frames = CalcNumFrames(&m_Frame);
						m_pInfo->tpf = m_Frame.ComputeTPF();
						m_nLength =
							(int)(m_pInfo->num_frames * m_pInfo->tpf * 1000);

						// PATRICK: fix
						// mpg123_ip.set_info(mpg123_title, m_nLength, m_nBitRate * 1000, m_nFrequency, m_nStereo);
					}
				}
				else
					nDispCount--;
				PlayFrame(&m_Frame);				
			}
			else
			{
				m_pOutput->BufferFreeAmt();
				m_pOutput->BufferFreeAmt();
				m_pInfo->eof = TRUE;
				Thread::usleep(10000);
			}
		}
		else
		{
			Thread::usleep(10000);
		}
	}
	// g_free(mpg123_title);
	// mpg123_title = NULL;
	StreamClose();
	if (bOutputOpened && !m_bAudioError) {
		m_pOutput->Close();
	}
	// g_free(m_pPCMSample);
	delete m_pPCMSample;
	m_pPCMSample = NULL;
	// g_free(filename);

	PlayComplete();
}

//------------------------------------------------------------------------------

void
PlayerMPG::ReadFrameInit()
{
	m_BSBuffer.SpaceReset();
	m_pInfo->output_audio = FALSE;
}

//------------------------------------------------------------------------------

void
PlayerMPG::OpenStream()
{
	if ((m_pFile = fopen(m_strFilename.c_str(), "rb")) != NULL) {
		if (StreamInit() == -1) {
			EM::I()->Trace() << "stream init failed" << endl;
			m_pInfo->eof = 1;
		}
	}
	else {
		EM::I()->Trace() << "file open failed" << endl;
		m_pInfo->eof = 1;
	}
}

//------------------------------------------------------------------------------

int
PlayerMPG::StreamInit()
{
	long nLength = GetFileInfo();

	if (nLength < 0) {
		return -1;
	}

	return 0;
}

//------------------------------------------------------------------------------

int
PlayerMPG::GetFileInfo()
{
	if (m_pFile == NULL) {
		return -1;
	}

	if (fseek(m_pFile, 0, SEEK_END) < 0) {
		return -1;
	}

	m_pInfo->filesize = ftell(m_pFile);

	if (fseek(m_pFile, -128, SEEK_END) < 0) {
		return -1;
	}

	guchar buf[3];
	if (FullRead(m_pFile, buf, 3) != 3) {
		return -1;
	}

	if (!strncmp((char *)buf, "TAG", 3)) {
		m_pInfo->filesize -= 128;
	}

	if (fseek(m_pFile, 0, SEEK_SET) < 0) {
		return -1;
	}

	if (m_pInfo->filesize <= 0) {
		return -1;
	}

	return m_pInfo->filesize;
}

//------------------------------------------------------------------------------

int
PlayerMPG::FullRead(FILE* pFile, unsigned char* pBuf, int nCount)
{
	int ret = -1;
	int cnt = 0;

	while (cnt < nCount) {
		if (pFile) {
			ret = fread(pBuf + cnt, 1, nCount - cnt, pFile);
		}

		if (ret < 0) {
			return ret;
		}

		if (ret == 0) {
			break;
		}

		cnt += ret;
	}

	return cnt;
}

//------------------------------------------------------------------------------

bool
PlayerMPG::ReadFrame(Frame* pFrame)
{
	m_BSBuffer.SetOldSize(pFrame->framesize);

	unsigned long newhead;
	if (!StreamHeadRead(&newhead)) {
		return false;
	}

	int nSSize = Frame::DecodeHeader(pFrame, newhead);
	m_BSBuffer.SetSize(nSSize);
	if (!HeadCheck(newhead) || (nSSize == 0)) {
		int nTry = 0;

		do {

			nTry++;

			if ((newhead & 0xffffff00) == ('I' << 24) + ('D' << 16) + ('3' << 8))
			{
				ReadID3v2Tag(newhead);
				if (!StreamHeadRead(&newhead)) {
					return false;
				}
			}
			else if (!StreamHeadShift(&newhead)) {
				return false;
			}

			nSSize = Frame::DecodeHeader(pFrame, newhead);
			m_BSBuffer.SetSize(nSSize);
		} while ((!HeadCheck(newhead) || (nSSize == 0)) &&
				 nTry < (256 * 1024));

		if (nTry >= (256 * 1024)) {
			return false;
		}

		m_pInfo->filesize -= nTry;
	}

	m_BSBuffer.FlipInit();

	if (!StreamReadFrameBody(m_BSBuffer.Get(), pFrame->framesize)) {
		return false;
	}

	m_BSBuffer.Reset();
	
	return true;
}

//------------------------------------------------------------------------------

bool
PlayerMPG::StreamHeadRead(unsigned long* newhead)
{
	unsigned char hbuf[4];

	if (FullRead(m_pFile, hbuf, 4) != 4)
		return false;

	*newhead = ((unsigned long) hbuf[0] << 24) |
		((unsigned long) hbuf[1] << 16) |
		((unsigned long) hbuf[2] << 8) |
		(unsigned long) hbuf[3];

	return true;
}

//------------------------------------------------------------------------------

bool
PlayerMPG::HeadCheck(unsigned long head)
{
	if ((head & 0xffe00000) != 0xffe00000)
		return false;
	if (!((head >> 17) & 3))
		return false;
	if (((head >> 12) & 0xf) == 0xf)
		return false;
	if (!((head >> 12) & 0xf))
		return false;
	if (((head >> 10) & 0x3) == 0x3)
		return false;
	if (((head >> 19) & 1) == 1 && ((head >> 17) & 3) == 3 && ((head >> 16) & 1) == 1)
		return false;
	if ((head & 0xffff0000) == 0xfffe0000)
		return false;
	
	return true;
}	

//------------------------------------------------------------------------------

bool
PlayerMPG::ReadID3v2Tag(unsigned long head)
{
	// guint8 version = 0;
	guint8 revision;
	guint8 flags;
	unsigned long headsize;
	
	// FullRead(m_pFile, (unsigned char*) &version, 1); 
	FullRead(m_pFile, (unsigned char*) &revision, 1); 
	FullRead(m_pFile, (unsigned char*) &flags, 1); 
	StreamHeadRead(&headsize);

	/*
	 * A invalid header could fool us into requesting insane
	 * amounts of memory.  Make sure the header size is
	 * reasonable.
	 */
	if ((m_pInfo->filesize && headsize > m_pInfo->filesize) ||
	    (!m_pInfo->filesize && headsize > 1000000)) {
		EM::I()->Warn() << "id3v2 -- insane headsize: " << headsize << endl;
		return false;
	}

	// if (mpg123_cfg.disable_id3v2)
	if (true)
	{
		// PATRICK: should subtract stuff already read from size???
		guint8 *tmp = (guint8 *)g_malloc(headsize);
		bool ret = (FullRead(m_pFile, tmp, headsize) == headsize);
		g_free(tmp);
		return ret;
	}

	
	// id3buf = g_malloc(hdrsize + sizeof (id3header));
	// memcpy(id3buf, &id3header, sizeof (id3header));

	/*
	 * Read ID3tag body.
	 */
	/*
	if (fullread(filept, id3buf + sizeof (id3header),
		     hdrsize) != hdrsize)
	{
		g_free(id3buf);
		return FALSE;
	}
	*/

	/*
	 * Get info from tag.
	 */
	/*
	id3d = id3_open_mem(id3buf, 0);
	mpg123_get_id3v2(id3d, &tag);
	if(!m_pInfo->first_frame)
	{
		songname = mpg123_title;
		mpg123_title = mpg123_format_song_title(&tag, mpg123_filename);
		mpg123_ip.set_info(mpg123_title, m_nLength,
				   m_nBitRate * 1000,
				   m_nFrequency, m_nStereo);
		if(songname)
			g_free(songname);
	}
	else
	{
		
		mpg123_title = mpg123_format_song_title(&tag, mpg123_filename);
	}
	id3_close(id3d);
	g_free(id3buf);
	*/
	
	return true;
}

//------------------------------------------------------------------------------

int
PlayerMPG::StreamHeadShift(unsigned long* pHead)
{
	unsigned char hbuf;

	if (FullRead(m_pFile, &hbuf, 1) != 1)
		return 0;
	
	*pHead <<= 8;
	*pHead |= hbuf;
	*pHead &= 0xffffffff;
	return 1;

}

//------------------------------------------------------------------------------

int
PlayerMPG::StreamReadFrameBody(unsigned char* buf, int size)
{
	long l;

	if ((l = FullRead(m_pFile, buf, size)) != size)
	{
		if (l <= 0)
			return 0;
		memset(buf + l, 0, size - l);
	}
	return 1;
}

//------------------------------------------------------------------------------

int
PlayerMPG::CalcNumFrames(Frame* fr)
{
	return (int) (m_pInfo->filesize / fr->ComputeBPF());
}

//------------------------------------------------------------------------------

void
PlayerMPG::PlayFrame(Frame* fr)
{
	if (fr->error_protection)
	{
		m_BSBuffer.IncrWordPointer(2);
		/*  mpg123_getbits(16); */	/* skip crc */
	}

	if (m_pLayer == NULL || m_pLayer->GetLevel() != fr->lay) {
		delete m_pLayer;
		if (fr->lay == 1) {
			m_pLayer = new Layer1(m_pInfo, m_pOutput, m_pVisualization,
								  m_pPCMSample, &m_nPCMPoint, &m_BSBuffer);
		}
		else if (fr->lay == 2) {
			m_pLayer = new Layer2(m_pInfo, m_pOutput, m_pVisualization,
								  m_pPCMSample, &m_nPCMPoint, &m_BSBuffer);
		}
		else if (fr->lay == 3) {
			m_pLayer = new Layer3(m_pInfo, m_pOutput, m_pVisualization,
								  m_pPCMSample, &m_nPCMPoint, &m_BSBuffer);
			m_pLayer->Init(fr->down_sample_sblimit);
		}
		else {
			EM::I()->Warn() << "unknown layer: " << fr->lay << endl;
			return;
		}
	}
	
	int nResult = m_pLayer->Run(fr);

	if (!nResult) {
		m_nSkipFrames = 2;
		m_pInfo->output_audio = FALSE;
	}
	else
	{
		if(!m_nSkipFrames) {
			m_pInfo->output_audio = TRUE;
		}
		else
			m_nSkipFrames--;
	}
}

//------------------------------------------------------------------------------

void
PlayerMPG::StreamJumpToFrame(Frame* fr, int nFrame)
{
	ReadFrameInit();
	fseek(m_pFile, nFrame * (fr->framesize + 4), SEEK_SET);
	ReadFrame(fr);
}

//------------------------------------------------------------------------------

void
PlayerMPG::StreamJumpToByte(Frame* fr, int nByte)
{
	fseek(m_pFile, nByte, SEEK_SET);
	ReadFrame(fr);
}

//------------------------------------------------------------------------------
// SeekPoint Copyright 1998 Xing Technology Corp.
//

int
PlayerMPG::SeekPoint(unsigned char TOC[100], int file_bytes, float percent)
{
	/* interpolate in TOC to get file seek point in bytes */ 
	int a, seekpoint;
	float fa, fb, fx;
                
	if (percent < 0.0f)
		percent = 0.0f;
        
	if (percent > 100.0f)
		percent = 100.0f;
	
	a = (int) percent;
	
	if (a > 99)
		a = 99;
	fa = TOC[a];
	
	if (a < 99)
	{
		fb = TOC[a + 1];
	}
	else
	{
		fb = 256.0f;
	}
        
	fx = fa + (fb - fa) * (percent - a);
	seekpoint = (int) ((1.0f / 256.0f) * fx * file_bytes);
	return seekpoint;
}

//------------------------------------------------------------------------------

void
PlayerMPG::StreamClose()
{

	if (m_pFile) {
		fclose(m_pFile);
	}
	else {
		// mpg123_http_close()
	}
}

//------------------------------------------------------------------------------

void
PlayerMPG::SetSynthFunctions(Frame* fr)
{
	delete fr->synth;
	fr->synth = new Synth1to1();

	delete fr->synth_mono;
	// fr->synth_mono = new Synth_1to1_mono();
	fr->synth_mono = NULL;
}

//------------------------------------------------------------------------------

void
PlayerMPG::Dump()
{
	EM::I()->Trace() << "PlayerMPG::Dump():" << endl;
	EM::I()->Trace() << "Filename: " << m_strFilename << endl;
	EM::I()->Trace() << "Bitrate: " << m_nBitRate << endl;
	EM::I()->Trace() << "Samplerate: " <<
		Frame::s_Freqs[m_Frame.sampling_frequency] << endl;

	std::string strMode;
	switch(m_nMode) {
	case 0:
		strMode = "Stereo";
		break;
	case 1:
		strMode = "Joint Stereo";
		break;
	case 2:
		strMode = "Dual Channel";
		break;
	case 3:
		strMode = "Single Channel";
		break;
	};
	EM::I()->Trace() << "Mode: " << strMode << endl;

	EM::I()->Trace() << "Error protection: " <<
		BoolStr(m_Frame.error_protection) << endl;
	EM::I()->Trace() << "Original: " << BoolStr(m_Frame.original) << endl;
	EM::I()->Trace() << "Copyright: " << BoolStr(m_Frame.copyright) << endl;
	
	AssertNotNull(m_pInfo);
	EM::I()->Trace() << "Frames: " << m_pInfo->num_frames << endl;
	EM::I()->Trace() << "Frame size: " << m_Frame.framesize << endl;
	
}

//------------------------------------------------------------------------------

std::string
PlayerMPG::BoolStr(bool bIn)
{
	if (bIn) {
		return "Yes";
	}
	else {
		return "No";
	}
}

//------------------------------------------------------------------------------

extern "C" {
	Player* maker()
	{
		return new PlayerMPG;
	}
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------

