// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// Synth.cpp
//
// $Id: Synth.cpp,v 1.1 2002/01/25 05:23:27 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "Synth.h"

#include <math.h>

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

float Synth::cos64[16];
float Synth::cos32[8];
float Synth::cos16[4];
float Synth::cos8[2];
float Synth::cos4[1];
float* Synth::s_apPoints[5] = {Synth::cos64,
							   Synth::cos32,
							   Synth::cos16,
							   Synth::cos8,
							   Synth::cos4};

const long Synth::s_nOutscale = 32768;

const long Synth::s_nWinBase[257] = {
	0, -1, -1, -1, -1, -1, -1, -2, -2, -2,
	-2, -3, -3, -4, -4, -5, -5, -6, -7, -7,
	-8, -9, -10, -11, -13, -14, -16, -17, -19, -21,
	-24, -26, -29, -31, -35, -38, -41, -45, -49, -53,
	-58, -63, -68, -73, -79, -85, -91, -97, -104, -111,
	-117, -125, -132, -139, -147, -154, -161, -169, -176, -183,
	-190, -196, -202, -208, -213, -218, -222, -225, -227, -228,
	-228, -227, -224, -221, -215, -208, -200, -189, -177, -163,
	-146, -127, -106, -83, -57, -29, 2, 36, 72, 111,
	153, 197, 244, 294, 347, 401, 459, 519, 581, 645,
	711, 779, 848, 919, 991, 1064, 1137, 1210, 1283, 1356,
	1428, 1498, 1567, 1634, 1698, 1759, 1817, 1870, 1919, 1962,
	2001, 2032, 2057, 2075, 2085, 2087, 2080, 2063, 2037, 2000,
	1952, 1893, 1822, 1739, 1644, 1535, 1414, 1280, 1131, 970,
	794, 605, 402, 185, -45, -288, -545, -814, -1095, -1388,
	-1692, -2006, -2330, -2663, -3004, -3351, -3705, -4063, -4425, -4788,
	-5153, -5517, -5879, -6237, -6589, -6935, -7271, -7597, -7910, -8209,
	-8491, -8755, -8998, -9219, -9416, -9585, -9727, -9838, -9916, -9959,
	-9966, -9935, -9863, -9750, -9592, -9389, -9139, -8840, -8492, -8092,
	-7640, -7134, -6574, -5959, -5288, -4561, -3776, -2935, -2037, -1082,
	-70, 998, 2122, 3300, 4533, 5818, 7154, 8540, 9975, 11455,
	12980, 14548, 16155, 17799, 19478, 21189, 22929, 24694, 26482, 28289,
	30112, 31947, 33791, 35640, 37489, 39336, 41176, 43006, 44821, 46617,
	48390, 50137, 51853, 53534, 55178, 56778, 58333, 59838, 61289, 62684,
	64019, 65290, 66494, 67629, 68692, 69679, 70590, 71420, 72169, 72835,
	73415, 73908, 74313, 74630, 74856, 74992, 75038};



//------------------------------------------------------------------------------

Synth::Synth()
{
	float* pCosTable;

	for (int i = 0; i < 5; i++) {
		int kr = 0x10 >> i;
		int divv = 0x40 >> i;
		pCosTable = s_apPoints[i];
		for (int k = 0; k < kr; k++) {
			pCosTable[k] = 1.0 / (2.0 * cos(M_PI * ((double) k * 2.0 + 1.0) /
											(double) divv));
		}
	}

	long nScaleVal = s_nOutscale;
	float* pTable = m_apDecwin;
	nScaleVal = -nScaleVal;
	int i, j;
	for (i = 0, j = 0; i < 256; i++, j++, pTable += 32) {
		if (pTable < m_apDecwin + 512 + 16) {
			pTable[16] = pTable[0] =
				(double) s_nWinBase[j] / 65536.0 * (double) nScaleVal;
		}
		if (i % 32 == 31) {
			pTable -= 1023;
		}
		if (i % 64 == 63) {
			nScaleVal = -nScaleVal;
		}
	}

	for (/* i = 256 */; i < 512; i++, j--, pTable += 32) {
		if (pTable < m_apDecwin + 512 + 16) {
			pTable[16] = pTable[0] = 
				(double) s_nWinBase[j] / 65536.0 * (double) nScaleVal;
		}
		if (i % 32 == 31) {
			pTable -= 1023;
		}
		if (i % 64 == 63) {
			nScaleVal = -nScaleVal;
		}
	}

}

//------------------------------------------------------------------------------

Synth::~Synth()
{

}

//------------------------------------------------------------------------------

void
Synth::mpg123_dct64(float * out0, float * out1, float * samples)
{
	float bufs[64];

	{
		register int i, j;
		register float *b1, *b2, *bs, *costab;

		b1 = samples;
		bs = bufs;
		costab = s_apPoints[0] + 16;
		b2 = b1 + 32;

		for (i = 15; i >= 0; i--)
			*bs++ = (*b1++ + *--b2);
		for (i = 15; i >= 0; i--)
			*bs++ = (*--b2 - *b1++) * *--costab;

		b1 = bufs;
		costab = s_apPoints[1] + 8;
		b2 = b1 + 16;

		{
			for (i = 7; i >= 0; i--)
				*bs++ = (*b1++ + *--b2);
			for (i = 7; i >= 0; i--)
				*bs++ = (*--b2 - *b1++) * *--costab;
			b2 += 32;
			costab += 8;
			for (i = 7; i >= 0; i--)
				*bs++ = (*b1++ + *--b2);
			for (i = 7; i >= 0; i--)
				*bs++ = (*b1++ - *--b2) * *--costab;
			b2 += 32;
		}

		bs = bufs;
		costab = s_apPoints[2];
		b2 = b1 + 8;

		for (j = 2; j; j--)
		{
			for (i = 3; i >= 0; i--)
				*bs++ = (*b1++ + *--b2);
			for (i = 3; i >= 0; i--)
				*bs++ = (*--b2 - *b1++) * costab[i];
			b2 += 16;
			for (i = 3; i >= 0; i--)
				*bs++ = (*b1++ + *--b2);
			for (i = 3; i >= 0; i--)
				*bs++ = (*b1++ - *--b2) * costab[i];
			b2 += 16;
		}

		b1 = bufs;
		costab = s_apPoints[3];
		b2 = b1 + 4;

		for (j = 4; j; j--)
		{
			*bs++ = (*b1++ + *--b2);
			*bs++ = (*b1++ + *--b2);
			*bs++ = (*--b2 - *b1++) * costab[1];
			*bs++ = (*--b2 - *b1++) * costab[0];
			b2 += 8;
			*bs++ = (*b1++ + *--b2);
			*bs++ = (*b1++ + *--b2);
			*bs++ = (*b1++ - *--b2) * costab[1];
			*bs++ = (*b1++ - *--b2) * costab[0];
			b2 += 8;
		}
		bs = bufs;
		costab = s_apPoints[4];

		for (j = 8; j; j--)
		{
			float v0, v1;

			v0 = *b1++;
			v1 = *b1++;
			*bs++ = (v0 + v1);
			*bs++ = (v0 - v1) * (*costab);
			v0 = *b1++;
			v1 = *b1++;
			*bs++ = (v0 + v1);
			*bs++ = (v1 - v0) * (*costab);
		}

	}

	{
		register float *b1;
		register int i;

		for (b1 = bufs, i = 8; i; i--, b1 += 4)
			b1[2] += b1[3];

		for (b1 = bufs, i = 4; i; i--, b1 += 8)
		{
			b1[4] += b1[6];
			b1[6] += b1[5];
			b1[5] += b1[7];
		}

		for (b1 = bufs, i = 2; i; i--, b1 += 16)
		{
			b1[8] += b1[12];
			b1[12] += b1[10];
			b1[10] += b1[14];
			b1[14] += b1[9];
			b1[9] += b1[13];
			b1[13] += b1[11];
			b1[11] += b1[15];
		}
	}

	out0[0x10 * 16] = bufs[0];
	out0[0x10 * 15] = bufs[16 + 0] + bufs[16 + 8];
	out0[0x10 * 14] = bufs[8];
	out0[0x10 * 13] = bufs[16 + 8] + bufs[16 + 4];
	out0[0x10 * 12] = bufs[4];
	out0[0x10 * 11] = bufs[16 + 4] + bufs[16 + 12];
	out0[0x10 * 10] = bufs[12];
	out0[0x10 * 9] = bufs[16 + 12] + bufs[16 + 2];
	out0[0x10 * 8] = bufs[2];
	out0[0x10 * 7] = bufs[16 + 2] + bufs[16 + 10];
	out0[0x10 * 6] = bufs[10];
	out0[0x10 * 5] = bufs[16 + 10] + bufs[16 + 6];
	out0[0x10 * 4] = bufs[6];
	out0[0x10 * 3] = bufs[16 + 6] + bufs[16 + 14];
	out0[0x10 * 2] = bufs[14];
	out0[0x10 * 1] = bufs[16 + 14] + bufs[16 + 1];
	out0[0x10 * 0] = bufs[1];

	out1[0x10 * 0] = bufs[1];
	out1[0x10 * 1] = bufs[16 + 1] + bufs[16 + 9];
	out1[0x10 * 2] = bufs[9];
	out1[0x10 * 3] = bufs[16 + 9] + bufs[16 + 5];
	out1[0x10 * 4] = bufs[5];
	out1[0x10 * 5] = bufs[16 + 5] + bufs[16 + 13];
	out1[0x10 * 6] = bufs[13];
	out1[0x10 * 7] = bufs[16 + 13] + bufs[16 + 3];
	out1[0x10 * 8] = bufs[3];
	out1[0x10 * 9] = bufs[16 + 3] + bufs[16 + 11];
	out1[0x10 * 10] = bufs[11];
	out1[0x10 * 11] = bufs[16 + 11] + bufs[16 + 7];
	out1[0x10 * 12] = bufs[7];
	out1[0x10 * 13] = bufs[16 + 7] + bufs[16 + 15];
	out1[0x10 * 14] = bufs[15];
	out1[0x10 * 15] = bufs[16 + 15];

}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
