// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// OutputOgg.cpp
//
// $Id: OutputOgg.cpp,v 1.8 2002/02/26 19:30:27 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "OutputOgg.h"

#include <string>
#include <vorbis/codec.h>
#include <vorbis/vorbisenc.h>
#include <xb/ErrorManager.h>
#include <xb/Filename.h>
#include <xb/String.h>

#include "PropList.h"
#include "System.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

OutputOgg::OutputOgg() :
	Output(),
	m_pFile(NULL),
	m_nFormat(FMT_S16_NE),
	m_nWritten(0),
	m_nChannels(0)
{

}

//------------------------------------------------------------------------------

OutputOgg::~OutputOgg()
{
	// delete m_pLameFlags?
}

//------------------------------------------------------------------------------

void
OutputOgg::Init()
{

}

//------------------------------------------------------------------------------

bool
OutputOgg::Open(AFormat nFormat, int nRate, int nChannels, int nTrack)
{
	m_nWritten = 0;
	m_nChannels = nChannels;
	
	xb::Filename strBasePath = System::GetPropList()->GetValue("SpoolRoot");
	std::string strFilename = strBasePath + "track" + xb::String(nTrack) +
		GetFileExtension();
	EM::I()->Trace() << "ogg filename: " << strFilename << endl;

	m_pFile = fopen(strFilename.c_str(), "wb+");

	if (m_pFile == NULL) {
		return false;
	}

	vorbis_info_init(&m_VI);

	// 0 = worst, 10 = best (opposite of lame...)
	int nQuality = System::GetPropList()->GetIntValue("EncoderQuality") / 10;
	if (nQuality < 0) { nQuality = 0; }
	if (nQuality > 10) { nQuality = 10; }
	EM::I()->Trace() << "ogg encoder quality: " << nQuality << endl;

	
	// vorbis encoder init
	//
	if (vorbis_encode_init_vbr(&m_VI, nChannels, nRate, nQuality)) {
		EM::I()->Error() << "vorbis encode init failed." << endl;
		vorbis_info_clear(&m_VI);
		return false;
	}

	// init analysis engine, stream encoder
	//

	vorbis_analysis_init(&m_VD, &m_VI);
	vorbis_block_init(&m_VD, &m_VB);

	ogg_stream_init(&m_OS, nTrack);

	// build 3 header packets
	//
	ogg_packet header_main;
	ogg_packet header_comments;
	ogg_packet header_codebooks;
	vorbis_comment vc;

	vorbis_comment_init(&vc);
	
	vorbis_analysis_headerout(&m_VD, &vc, &header_main, &header_comments,
							  &header_codebooks);

	// stream header packets
	//
	ogg_stream_packetin(&m_OS, &header_main);
	ogg_stream_packetin(&m_OS, &header_comments);
	ogg_stream_packetin(&m_OS, &header_codebooks);

	ogg_page og;
	
	int nResult = ogg_stream_flush(&m_OS, &og);
	
	while (nResult > 0) {

		int nOut = WritePage(&og);
		if (nOut != og.header_len + og.body_len) {
			EM::I()->Error() << "failed writing header to output stream" << endl;
			return false;
		}

		m_nWritten += nOut;
		
		nResult = ogg_stream_flush(&m_OS, &og);
	}
	
	return true;
}

//------------------------------------------------------------------------------

void
OutputOgg::Write(void* pData, int nLength)
{

	AssertNotNull(m_pFile);

	float **buffer = vorbis_analysis_buffer(&m_VD, 1024);
	int nBytesPerSample = 2;
	int nRealSamples = nLength / (nBytesPerSample * m_nChannels);

	/*
	char* pCData = (char *)pData;
	for (int i = 0; i < nRealSamples; i++) {
		for (int j = 0; j < m_nChannels; j++) {
			buffer[j][i] = ((pCData[i * 2 * m_nChannels + 2 * j] << 8) |
							(pCData[i * 2 * m_nChannels + 2 * j + 1] & 0xff)) /
				32768.0f;
		}
	}
	*/
	gint16* pCData = (gint16 *)pData;
	for (int i = 0; i < nRealSamples; i++) {
		for (int j = 0; j < m_nChannels; j++) {
			buffer[j][i] = pCData[i * m_nChannels] / 32768.0f;
		}
	}
	
	vorbis_analysis_wrote(&m_VD, nRealSamples);

	ogg_packet op;
	
	while (vorbis_analysis_blockout(&m_VD, &m_VB) == 1) {

		vorbis_analysis(&m_VB, NULL);
		vorbis_bitrate_addblock(&m_VB);

		while (vorbis_bitrate_flushpacket(&m_VD, &op)) {
			ogg_stream_packetin(&m_OS, &op);

			bool bDone = false;

			while (!bDone) {
				ogg_page og;
				int nResult = ogg_stream_pageout(&m_OS, &og);
				if (!nResult) break;

				int nOut = WritePage(&og);
				if (nOut != og.header_len + og.body_len) {
					EM::I()->Error() <<"Failed writing data to output stream" <<
						endl;
					return;
				}

				m_nWritten += nOut;

				if (ogg_page_eos(&og)) {
					bDone = true;
				}
			}
		}
	}
}

//------------------------------------------------------------------------------

void
OutputOgg::Close()
{
	ogg_stream_clear(&m_OS);

	vorbis_info_clear(&m_VI);
	vorbis_block_clear(&m_VB);
	vorbis_dsp_clear(&m_VD);

	if (m_pFile) {

		/*
		// flush
		// 
		unsigned char pBuffer[LAME_MAXMP3BUFFER];
		int nSize = lame_encode_flush(m_pLameFlags, pBuffer, sizeof(pBuffer));

		if (nSize < 0) {
			// errror
		}

		fwrite(pBuffer, 1, nSize, m_pFile);

		*/
		
		fclose(m_pFile);

		m_nWritten = 0;
	}

	m_pFile = NULL;
}

//------------------------------------------------------------------------------

int
OutputOgg::ElapsedTime()
{
	int nResult = 0;

	/*
	if (m_Header.byte_p_sec != 0.0) {
		nResult = (m_nWritten * 1000) / m_Header.byte_p_sec;
	}
	*/
	
	return nResult;
}

//------------------------------------------------------------------------------
/*
void
OutputOgg::ConvertBuffer(gpointer buffer, gint length)
{
	gint i;

	if (m_nFormat == FMT_S8)
	{
		guint8 *ptr1 = (guint8 *)buffer;
		gint8 *ptr2 = (gint8 *)buffer;

		for (i = 0; i < length; i++)
			*(ptr1++) = *(ptr2++) ^ 128;
	}
	if (m_nFormat == FMT_S16_BE)
	{
		gint16 *ptr = (gint16 *)buffer;

		for (i = 0; i < length >> 1; i++, ptr++)
			*ptr = GUINT16_SWAP_LE_BE(*ptr);
	}
	if (m_nFormat == FMT_S16_NE)
	{
		gint16 *ptr = (gint16 *)buffer;

		for (i = 0; i < length >> 1; i++, ptr++)
			*ptr = GINT16_TO_LE(*ptr);
	}
	if (m_nFormat == FMT_U16_BE)
	{
		gint16 *ptr1 = (gint16 *)buffer;
		guint16 *ptr2 = (guint16 *)buffer;

		for (i = 0; i < length >> 1; i++, ptr2++)
			*(ptr1++) = GINT16_TO_LE(GUINT16_FROM_BE(*ptr2) ^ 32768);
	}
	if (m_nFormat == FMT_U16_LE)
	{
		gint16 *ptr1 = (gint16 *)buffer;
		guint16 *ptr2 = (guint16 *)buffer;

		for (i = 0; i < length >> 1; i++, ptr2++)
			*(ptr1++) = GINT16_TO_LE(GUINT16_FROM_LE(*ptr2) ^ 32768);
	}
	if (m_nFormat == FMT_U16_NE)
	{
		gint16 *ptr1 = (gint16 *)buffer;
		guint16 *ptr2 = (guint16 *)buffer;

		for (i = 0; i < length >> 1; i++, ptr2++)
			*(ptr1++) = GINT16_TO_LE((*ptr2) ^ 32768);
	}

}
*/
//------------------------------------------------------------------------------

int
OutputOgg::WritePage(ogg_page* pPage)
{
	AssertNotNull(m_pFile);
	AssertNotNull(pPage);
	
	int nResult = 0;

	nResult = fwrite(pPage->header, 1, pPage->header_len, m_pFile);
	nResult += fwrite(pPage->body, 1, pPage->body_len, m_pFile);

	return nResult;
}

//------------------------------------------------------------------------------

extern "C" {
	Output* maker()
	{
		return new OutputOgg;
	}
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
