// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// OutputOSS.cpp
//
// $Id: OutputOSS.cpp,v 1.11 2002/02/26 19:30:27 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------
//
// this whole file is a heavily modified version of the OSS plugin for xmms
//
//------------------------------------------------------------------------------

#include <config.h>

#include "OutputOSS.h"

#include <algorithm>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

#ifdef HAVE_SYS_SOUNDCARD_H
#include <sys/soundcard.h>
#elif defined(HAVE_MACHINE_SOUNDCARD_H)
#include <machine/soundcard.h>
#endif

#include <xb/ErrorManager.h>

#include "PropList.h"
#include "System.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

const int OutputOSS::s_nFrags = 32;
const int OutputOSS::s_nBufferSize = 3000;
const int OutputOSS::s_nPreBufferSize = 25;

//------------------------------------------------------------------------------

OutputOSS::OutputOSS() :
	Output(),
	m_pBuffer(NULL),
	m_bSuspend(false)
{

}

//------------------------------------------------------------------------------

OutputOSS::~OutputOSS()
{
	EM::I()->Trace() << "OutputOSS::~OutputOSS()" << endl;
	g_free(m_pBuffer);
	m_pBuffer = NULL;
}

//------------------------------------------------------------------------------

void
OutputOSS::Init()
{
}

//------------------------------------------------------------------------------

void
OutputOSS::GetVolume(int& nLeft, int& nRight)
{
	m_Mixer.GetVolume(nLeft, nRight);
}

//------------------------------------------------------------------------------

void
OutputOSS::SetVolume(int nLeft, int nRight)
{
	m_Mixer.SetVolume(nLeft, nRight);
}

//------------------------------------------------------------------------------

bool
OutputOSS::Open(AFormat nFormat, int nRate, int nChannels, int nTrack)
{
	std::string strDevice = System::GetPropList()->GetValue("DspDevice");
	if (strDevice.length() > 0) {
		m_strDevice = strDevice;
	}
	else {
		m_strDevice = "/dev/dsp";
	}

	m_nFile = open(m_strDevice.c_str(), O_WRONLY);
	if (m_nFile == -1) {
		EM::I()->Warn() << "OutputOSS::Open(): failed to open audio device (" <<
			m_strDevice << "): " << strerror(errno) << endl;
		return false;
	}

	SetupFormat(nFormat, nRate, nChannels);

	m_nInputFormat = m_nFormat;
	m_nInputChannels = m_nChannels;
	m_nInputFrequency = m_nFrequency;
	m_nInputBPS = m_nBPS;
	
	if (!m_bRealtime) {
		m_nBufferSize = (s_nBufferSize * m_nInputBPS) / 1000;
		if (m_nBufferSize < 8192) {
			m_nBufferSize = 8192;
		}
		m_nPreBufferSize = (m_nBufferSize * s_nPreBufferSize) / 100;
		if (m_nBufferSize - m_nPreBufferSize < 4096) {
			m_nPreBufferSize = m_nBufferSize - 4096;
		}

		m_nBufferSize += m_nDeviceBufferSize;
		if (m_pBuffer != NULL) {
			g_free(m_pBuffer);
		}
		m_pBuffer = g_malloc0(m_nBufferSize);
	}

	
	
	m_nFlush = -1;
	
	m_bPreBuffer = true;
	m_bPaused = false;
	m_bDoPause = false;
	m_bUnPause = false;
	m_bRemovePreBuffer = false;

	m_nWriteIndex = 0;
	m_nReadIndex = 0;
	m_nOutputTimeOffset = 0;
	m_nWritten = 0;
	m_nOutputBytes = 0;

	SetAudioParams();
	m_bGoing = true;
	if (!m_bRealtime) {
		if (!m_bSuspend) {
			Start();
		}
		else {
			m_bSuspend = false;
			Resume();
		}
	}

	return true;
}

//------------------------------------------------------------------------------

void
OutputOSS::Write(void* pData, int nLength)
{
	if (!m_bRealtime) {
		int nOffset = 0;
		
		m_bRemovePreBuffer = false;

		m_nWritten += nLength;
		while (nLength > 0) {
			int nCount = min(nLength, m_nBufferSize - m_nWriteIndex);
			memcpy((gchar *)m_pBuffer + m_nWriteIndex,
				   (gchar *)pData + nOffset, nCount);
			m_nWriteIndex = (m_nWriteIndex + nCount) % m_nBufferSize;
			nLength -= nCount;
			nOffset += nCount;
		}
	}
	else {
		if (m_bPaused) {
			return;
		}

		m_nWritten += WriteAudio(pData, nLength);
	}
}

//------------------------------------------------------------------------------

void
OutputOSS::Close()
{
	if (!m_bGoing) {
		return;
	}

	/*
	m_bGoing = false;
	if (!m_bRealtime) {
		// PATRICK: Terminate()???
		Terminate();
		// pthread_join(buffer_thread, NULL);
		while (isRunning()) {
			EM::I()->Trace("OutputOSS:: waiting for thread to stop...");
			Thread::usleep(10000);
		}
	}
	else {
		ioctl(m_nFile, SNDCTL_DSP_RESET, 0);
		close(m_nFile);
	}
	*/

	// trying suspend instead...
	if (!m_bRealtime) {
		if (m_bSuspend) {
			return;
		}
		
		Suspend();
		m_bSuspend = true;
		close(m_nFile);
	}
	else {
		ioctl(m_nFile, SNDCTL_DSP_RESET, 0);
		close(m_nFile);
	}
	
	m_nWriteIndex = 0;
	m_nReadIndex = 0;

}

//------------------------------------------------------------------------------

void
OutputOSS::Flush(int nTime)
{
	if (!m_bRealtime) {
		m_nFlush = nTime;
		while (m_nFlush != -1) {
			Thread::usleep(10000);
		}
	}
	else {
		ioctl(m_nFile, SNDCTL_DSP_RESET, 0);
		close(m_nFile);
		m_nFile = open(m_strDevice.c_str(), O_WRONLY);
		SetAudioParams();
		m_nOutputTimeOffset = nTime;
		m_nWritten = (guint64)(nTime / 10) * (guint64)(m_nInputBPS / 100);		
		m_nOutputBytes = 0;
	}
}

//------------------------------------------------------------------------------

void
OutputOSS::Pause(bool bPause)
{
	if (!m_bRealtime) {
		if (bPause) {
			m_bDoPause = true;
		}
		else {
			m_bUnPause = true;
		}
	}
	else {
		m_bPaused = bPause;
	}
}

//------------------------------------------------------------------------------

bool
OutputOSS::BufferPlaying()
{
	if (!m_bGoing) {
		return false;
	}

	if (m_bRealtime) {
		CalcDeviceBufferUsed();
	}

	if (!Used() && (m_nDeviceBufferUsed - (3 * m_nBlockSize)) <= 0)
		return false;

	return true;
}

//------------------------------------------------------------------------------

int
OutputOSS::ElapsedTime()
{
	guint64 bytes;

	if (!m_nFile || !m_bGoing) {
		return 0;
	}

	if (m_bRealtime) {
		CalcDeviceBufferUsed();
	}

	bytes = max(0, m_nOutputBytes - m_nDeviceBufferUsed);

	return m_nOutputTimeOffset + (gint) ((bytes  * 1000) / m_nEBPS);
}

//------------------------------------------------------------------------------

guint64
OutputOSS::WrittenTime() const
{
	long nResult = 0;
	
	if (m_bGoing) {
		nResult = (m_nWritten * 1000) / m_nBPS;
	}

	return nResult;
}

//------------------------------------------------------------------------------

int
OutputOSS::BufferFreeAmt()
{
	int nResult = 0;
	
	if (!m_bRealtime) {
		
		if (m_bRemovePreBuffer && m_bPreBuffer) {
			m_bPreBuffer = false;
			m_bRemovePreBuffer = false;
		}

		if (m_bPreBuffer) {
			m_bRemovePreBuffer = true;
		}

		if (m_nReadIndex > m_nWriteIndex) {
			nResult = (m_nReadIndex - m_nWriteIndex) - m_nDeviceBufferSize - 1;
		}
		else {
			nResult = m_nBufferSize - (m_nWriteIndex - m_nReadIndex) -
											  m_nDeviceBufferSize - 1;
		}
	}
	else {
		if (m_bPaused) {
			nResult = 0;
		}
		else {
			nResult = 1000000;
		}
	}

	return nResult;
}

//------------------------------------------------------------------------------

void
OutputOSS::Run()
{
	int nLength;

	while (m_bGoing) {

		if (Used() > m_nPreBufferSize) {
			m_bPreBuffer = false;
		}

		if (Used() > 0  && !m_bPaused && !m_bPreBuffer) {
			struct timeval tv;
			fd_set set;
			tv.tv_sec = 0;
			tv.tv_usec = 10000;
			FD_ZERO(&set);
			FD_SET(m_nFile, &set);
			if (!m_bSelectWorks ||
				(select(m_nFile + 1, NULL, &set, NULL, &tv) > 0)) {
				
				nLength = min(m_nBlockSize, Used());

				while (nLength > 0) {
					int nCount = min(nLength, m_nBufferSize - m_nReadIndex);

					nCount =
						WriteAudio((gchar *)m_pBuffer + m_nReadIndex, nCount);
					m_nReadIndex = (m_nReadIndex + nCount) % m_nBufferSize;

					nLength -= nCount;
				}

				if (!Used()) {
					ioctl(m_nFile, SNDCTL_DSP_POST, 0);
				}
			}
		}
		else {
			Thread::usleep(10000);
		}

		CalcDeviceBufferUsed();

		if (m_bDoPause && !m_bPaused) {
			m_bDoPause = false;
			m_bPaused = true;
			m_nReadIndex -= m_nDeviceBufferUsed;
			m_nOutputBytes -= m_nDeviceBufferUsed;
			if (m_nReadIndex < 0) {
				m_nReadIndex += m_nBufferSize;
			}
			ioctl(m_nFile, SNDCTL_DSP_RESET, 0);
		}

		if (m_bUnPause && m_bPaused) {
			m_bUnPause = false;
			close(m_nFile);
			m_nFile = open(m_strDevice.c_str(), O_WRONLY);
			SetAudioParams();
			m_bPaused = false;
		}

		if (m_nFlush != -1) {
			ioctl(m_nFile, SNDCTL_DSP_RESET, 0);
			close(m_nFile);
			m_nFile = open(m_strDevice.c_str(), O_WRONLY);
			SetAudioParams();
			m_nOutputTimeOffset = m_nFlush;
			m_nWritten = (guint64)(m_nFlush / 10) * (guint64)(m_nInputBPS / 100);
			m_nReadIndex = 0;
			m_nWriteIndex = 0;
			m_nOutputBytes = 0;
			m_nFlush = -1;
			m_bPreBuffer = true;
		}
	}

	ioctl(m_nFile, SNDCTL_DSP_RESET, 0);
	close(m_nFile);
	g_free(m_pBuffer);
	m_pBuffer = NULL;

	Terminate();
}

//------------------------------------------------------------------------------

void
OutputOSS::SetupFormat(AFormat nFormat, int nRate, int nChannels)
{
	
	m_nFormat = nFormat;
	m_nFrequency = nRate;
	m_nChannels = nChannels;

	switch (nFormat)
	{
		case FMT_U8:
			m_nOSSFormat = AFMT_U8;
			break;
		case FMT_S8:
			m_nOSSFormat = AFMT_S8;
			break;
		case FMT_U16_LE:
			m_nOSSFormat = AFMT_U16_LE;
			break;
		case FMT_U16_BE:
			m_nOSSFormat = AFMT_U16_BE;
			break;
		case FMT_U16_NE:
#ifdef AFMT_U16_NE
			m_nOSSFormat = AFMT_U16_NE;
#else
#ifdef WORDS_BIGENDIAN
			m_nOSSFormat = AFMT_U16_BE;
#else
			m_nOSSFormat = AFMT_U16_LE;
#endif
#endif
			break;
		case FMT_S16_LE:
			m_nOSSFormat = AFMT_S16_LE;
			break;
		case FMT_S16_BE:
			m_nOSSFormat = AFMT_S16_BE;
			break;
		case FMT_S16_NE:
#ifdef AFMT_S16_NE
			m_nOSSFormat = AFMT_S16_NE;
#else
#ifdef WORDS_BIGENDIAN
			m_nOSSFormat = AFMT_S16_BE;
#else
			m_nOSSFormat = AFMT_S16_LE;
#endif
#endif
			break;
	}

	m_nBPS = m_nFrequency * m_nChannels;
	if (m_nOSSFormat == AFMT_U16_BE ||
		m_nOSSFormat == AFMT_U16_LE ||
		m_nOSSFormat == AFMT_S16_BE ||
		m_nOSSFormat == AFMT_S16_LE) {
		m_nBPS *= 2;
	}

	m_nFragSize = 0;
	while (1L << m_nFragSize < m_nBPS / 25) {
		m_nFragSize++;
	}
	m_nFragSize--;

	m_nDeviceBufferSize = ((1L << m_nFragSize) * (s_nFrags + 1));

}

//------------------------------------------------------------------------------

void
OutputOSS::SetAudioParams()
{
	ioctl(m_nFile, SNDCTL_DSP_RESET, 0);

	int nFrag = (s_nFrags << 16) | m_nFragSize;
	ioctl(m_nFile, SNDCTL_DSP_SETFRAGMENT, &nFrag);
	ioctl(m_nFile, SNDCTL_DSP_SETFMT, &m_nOSSFormat);
	ioctl(m_nFile, SNDCTL_DSP_SETFMT, &m_nOSSFormat);
	int nStereo = m_nChannels - 1;
	ioctl(m_nFile, SNDCTL_DSP_STEREO, &nStereo);
	m_nEFrequency = m_nFrequency;
	ioctl(m_nFile, SNDCTL_DSP_SPEED, &m_nEFrequency);

	if (ioctl(m_nFile, SNDCTL_DSP_GETBLKSIZE, &m_nBlockSize) == -1) {
		m_nBlockSize = 1L << m_nFragSize;
	}

	m_nEBPS = m_nEFrequency * m_nChannels;
	if (m_nOSSFormat == AFMT_U16_BE || m_nOSSFormat == AFMT_U16_LE ||
	    m_nOSSFormat == AFMT_S16_BE || m_nOSSFormat == AFMT_S16_LE) {
		m_nEBPS *= 2;
	}

	struct timeval tv;
	fd_set set;
	tv.tv_sec = 0;
	tv.tv_usec = 50000;
	FD_ZERO(&set);
	FD_SET(m_nFile, &set);
	int ret = select(m_nFile + 1, NULL, &set, NULL, &tv);
	if (ret > 0) {
		m_bSelectWorks = true;
	}
	else {
		m_bSelectWorks = false;
	}

}

//------------------------------------------------------------------------------

int
OutputOSS::WriteAudio(gpointer pData, int nLength)
{
	
	int nResult = 0;
	
	if (m_nInputFormat != m_nFormat || m_nInputFrequency != m_nFrequency ||
		m_nInputChannels != m_nChannels) {
		m_nOutputTimeOffset += (int)((m_nOutputBytes * 1000) / m_nEBPS);
		m_nOutputBytes = 0;
		SetupFormat(m_nInputFormat, m_nInputFrequency, m_nInputChannels);
		m_nFrequency = m_nInputFrequency;
		m_nChannels = m_nInputChannels;
		close(m_nFile);
		m_nFile = open(m_strDevice.c_str(), O_WRONLY);
		SetAudioParams();
	}

	audio_buf_info bufInfo;
	if (m_bRealtime && !ioctl(m_nFile, SNDCTL_DSP_GETOSPACE, &bufInfo)) {
		while (bufInfo.bytes < nLength) {
			Thread::usleep(10000);
			if (ioctl(m_nFile, SNDCTL_DSP_GETOSPACE, &bufInfo))
				break;
		}
	}

	if (m_nFrequency == m_nEFrequency) {
		char byte0[2];
		byte0[0] = ((char *)pData)[0];
		byte0[1] = '\0';
		// int n = ((char *)pData)[0];
		nResult = write(m_nFile, pData, nLength);
	}
	else {
		nResult = DownSample(pData, nLength, m_nFrequency, m_nEFrequency);
	}

	m_nOutputBytes += nResult;

	/* FIXME: Handle short writes when downsampling */
	if (m_nFrequency != m_nEFrequency) {
		nResult = nLength;
	}

	return nResult;
}

//------------------------------------------------------------------------------

void
OutputOSS::swap_words(guint16 *buffer, gint length)
{
	guint16 *ptr = buffer;
	gint i;
	for(i = 0; i < length; i++, ptr++)
		*ptr = ((*ptr & 0x00FF) << 8) | (*ptr >> 8);
}

//------------------------------------------------------------------------------

#define RESAMPLE_STEREO(sample_type)				\
do {								\
	const gint shift = sizeof (sample_type);		\
    gint i, in_samples, out_samples, x, delta;		\
	sample_type *inptr = (sample_type *)ob, *outptr;	\
	guint nlen = (((length >> shift) * espeed) / speed);	\
	if (nlen == 0)						\
		break;						\
	nlen <<= shift;						\
	if((big_endian &&					\
	    (m_nOSSFormat == AFMT_S16_LE ||			\
	     m_nOSSFormat == AFMT_U16_LE)) ||			\
	   (!big_endian &&					\
	    (m_nOSSFormat == AFMT_S16_BE ||			\
	     m_nOSSFormat == AFMT_U16_BE)))			\
		swap_words((guint16*)ob, length >> 1);			\
	if(nlen > nbuffer_size)					\
	{							\
		nbuffer = g_realloc(nbuffer, nlen);		\
		nbuffer_size = nlen;				\
	}							\
	outptr = (sample_type *)nbuffer;			\
	in_samples = length >> shift;				\
        out_samples = nlen >> shift;				\
	delta = (in_samples << 12) / out_samples;		\
	for (x = 0, i = 0; i < out_samples; i++)		\
	{							\
		gint x1, frac;					\
		x1 = (x >> 12) << 12;				\
		frac = x - x1;					\
		*outptr++ =					\
			(sample_type)				\
			((inptr[(x1 >> 12) << 1] *		\
			  ((1<<12) - frac) +			\
			  inptr[((x1 >> 12) + 1) << 1] *	\
			  frac) >> 12);				\
		*outptr++ =					\
			(sample_type)				\
			((inptr[((x1 >> 12) << 1) + 1] *	\
			  ((1<<12) - frac) +			\
			  inptr[(((x1 >> 12) + 1) << 1) + 1] *	\
			  frac) >> 12);				\
		x += delta;					\
	}							\
	if ((big_endian &&					\
	     (m_nOSSFormat == AFMT_S16_LE ||			\
	      m_nOSSFormat == AFMT_U16_LE)) ||			\
	    (!big_endian &&					\
	     (m_nOSSFormat == AFMT_S16_BE ||			\
	      m_nOSSFormat == AFMT_U16_BE)))			\
		swap_words((guint16*)nbuffer, nlen >> 1);			\
	w = write(m_nFile, nbuffer, nlen);				\
} while (0)

//------------------------------------------------------------------------------

#define RESAMPLE_MONO(sample_type)				\
do {								\
	const gint shift = sizeof (sample_type) - 1;		\
        gint i, x, delta, in_samples, out_samples;		\
	sample_type *inptr = (sample_type *)ob, *outptr;	\
	guint nlen = (((length >> shift) * espeed) / speed);	\
	if (nlen == 0)						\
		break;						\
	nlen <<= shift;						\
	if((big_endian &&					\
	    (m_nOSSFormat == AFMT_S16_LE ||			\
	     m_nOSSFormat == AFMT_U16_LE)) ||			\
	   (!big_endian &&					\
	    (m_nOSSFormat == AFMT_S16_BE ||			\
	     m_nOSSFormat == AFMT_U16_BE)))			\
		swap_words((guint16*)ob, length >> 1);			\
	if(nlen > nbuffer_size)					\
	{							\
		nbuffer = g_realloc(nbuffer, nlen);		\
		nbuffer_size = nlen;				\
	}							\
	outptr = (sample_type *)nbuffer;			\
	in_samples = length >> shift;				\
        out_samples = nlen >> shift;				\
	delta = ((length >> shift) << 12) / out_samples;	\
	for (x = 0, i = 0; i < out_samples; i++)		\
	{							\
		gint x1, frac;					\
		x1 = (x >> 12) << 12;				\
		frac = x - x1;					\
		*outptr++ =					\
			(sample_type)				\
			((inptr[x1 >> 12] * ((1<<12) - frac) +	\
			  inptr[(x1 >> 12) + 1] * frac) >> 12);	\
		x += delta;					\
	}							\
	if ((big_endian &&					\
	    (m_nOSSFormat == AFMT_S16_LE ||			\
	     m_nOSSFormat == AFMT_U16_LE)) ||			\
	   (!big_endian &&					\
	    (m_nOSSFormat == AFMT_S16_BE ||			\
	     m_nOSSFormat == AFMT_U16_BE)))			\
		swap_words((guint16*)nbuffer, nlen >> 1);			\
	w = write(m_nFile, nbuffer, nlen);				\
} while (0)

//------------------------------------------------------------------------------

int
OutputOSS::DownSample(gpointer ob, guint length, guint speed, guint espeed)
{
	guint w = 0;
	static gpointer nbuffer = NULL;
	static guint nbuffer_size = 0;
#ifdef WORDS_BIGENDIAN
	const gboolean big_endian = TRUE;
#else
	const gboolean big_endian = FALSE;
#endif

	switch (m_nOSSFormat)
	{
		case AFMT_S16_BE:
		case AFMT_S16_LE:
			if(m_nChannels == 2)
				RESAMPLE_STEREO(gint16);
			else
				RESAMPLE_MONO(gint16);
			break;
		case AFMT_U16_BE:
		case AFMT_U16_LE:
			if(m_nChannels == 2)
				RESAMPLE_STEREO(guint16);
			else
				RESAMPLE_MONO(guint16);
			break;
		case AFMT_S8:
			if(m_nChannels == 2)
				RESAMPLE_STEREO(gint8);
			else
				RESAMPLE_MONO(gint8);
			break;
		case AFMT_U8:
			if(m_nChannels == 2)
				RESAMPLE_STEREO(guint8);
			else
				RESAMPLE_MONO(guint8);
			break;
	}
	return w;
}

//------------------------------------------------------------------------------

int
OutputOSS::Used() const
{
	int nResult = 0;

	if (!m_bRealtime) {
		if (m_nWriteIndex >= m_nReadIndex) {
			nResult = m_nWriteIndex - m_nReadIndex;
		}
		else {
			nResult = m_nBufferSize - (m_nReadIndex - m_nWriteIndex);
		}
	}

	return nResult;
}

//------------------------------------------------------------------------------

void
OutputOSS::CalcDeviceBufferUsed()
{
	audio_buf_info bufInfo;
	if (m_bPaused) {
		m_nDeviceBufferUsed = 0;
	}
	else if (!ioctl(m_nFile, SNDCTL_DSP_GETOSPACE, &bufInfo)) {
		m_nDeviceBufferUsed  = (bufInfo.fragstotal * bufInfo.fragsize) -
			bufInfo.bytes;
	}
}

//------------------------------------------------------------------------------

extern "C" {
	Output* maker()
	{
		return new OutputOSS;
	}
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
