// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// AudioCD.cpp
//
// $Id: AudioCD.cpp,v 1.27 2002/03/09 19:14:08 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include <config.h>

#include "AudioCD.h"

extern "C" {
#include <cdda_interface.h>
}

#include <unistd.h>
#include <string.h>

#ifdef HAVE_LINUX_CDROM_H
extern "C" {
#include <linux/cdrom.h>
#include <sys/ioctl.h>	
}
#endif

#include <xb/ErrorManager.h>
#include <xb/String.h>

#include "ChoiceDialog.h"
#include "PropList.h"
#include "System.h"
#include "Util.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

const int AudioCD::s_nInvalidDevice;

//------------------------------------------------------------------------------

AudioCD::AudioCD() :
	m_pDiscData(NULL)
{
	m_bDataCurrent = false;
	std::string strDeviceName =	System::GetPropList()->GetValue("CDReader");

	m_nDevice = cd_init_device((char*)strDeviceName.c_str());

	if (m_nDevice == s_nInvalidDevice) {
		cerr << "couldn't init " << strDeviceName << endl;
	}

}

//------------------------------------------------------------------------------

AudioCD::~AudioCD()
{
	Stop();
	cd_finish(m_nDevice);
}

//------------------------------------------------------------------------------

void
AudioCD::Update()
{
	m_SongList.clear();

	if (!IsAudio()) {
		EM::I()->Trace() << "AudioCD::Update() -- not an audio cd" << endl;
		return;
	}
	
	GetCDDBData();

	std::string strArtist("(unknown)");
	std::string strAlbum("(unknown)");
	if (m_pDiscData.get() != NULL) {
		strArtist = m_pDiscData->data_artist;
		if (strArtist.length() == 0) {
			strArtist = "(unknown)";
		}
		strAlbum = m_pDiscData->data_title;
		if (strAlbum.length() == 0) {
			strAlbum = "(unknown)";
		}
	}

	struct disc_info info;
	cd_stat(m_nDevice, &info);

	if (info.disc_total_tracks > 100) {
		info.disc_total_tracks = 100;
	}
	
	for (int i = 0; i < info.disc_total_tracks; i++) {
		struct disc_timeval trackTime = info.disc_track[i].track_length;

		std::string strTime = Util::GetTimeStr(trackTime.minutes,
											   trackTime.seconds);
			
		std::string strTrackName = m_pDiscData->data_track[i].track_name;
		if (strTrackName.length() == 0) {
			strTrackName = "track " + xb::String(i + 1);
		}
			
		Song song(strArtist, strAlbum, strTrackName, xb::String(i + 1),
				  strTime, trackTime.minutes * 60 + trackTime.seconds);
		m_SongList.push_back(song);

		m_bDataCurrent = true;
	}

}

//------------------------------------------------------------------------------

const SongList&
AudioCD::GetSongList()
{
	EM::I()->Noise() << "device:  " << m_nDevice << endl;
	if (CheckDiscInDrive() &&
		((m_pDiscData.get() == NULL) ||
		 (m_pDiscData->data_id != (long long)cddb_discid(m_nDevice)) ||
		 (!m_bDataCurrent))) {
		
		Update();
	}

	return m_SongList;
}

//------------------------------------------------------------------------------

void
AudioCD::Dump() const
{
	// dump info
	AssertNotNull(m_pDiscData.get());
	
	EM::I()->Trace() << "Title: " << m_pDiscData->data_title << endl;
	EM::I()->Trace() << "Artist: " <<  m_pDiscData->data_artist << endl;
	EM::I()->Trace() << "Extended: " <<  m_pDiscData->data_extended << endl;

	for (int i = 0; i < MAX_TRACKS; i++) {
		if (strlen(m_pDiscData->data_track[i].track_name) > 0) {
			EM::I()->Trace() << i << ":  " <<
				m_pDiscData->data_track[i].track_name << 
				" (" << m_pDiscData->data_track[i].track_artist << ") [" <<
				m_pDiscData->data_track[i].track_extended << "]" << endl;
		}
	}
	
}

//------------------------------------------------------------------------------

void
AudioCD::GetCDDBData()
{
	m_pDiscData.reset(new disc_data);
	m_pDiscData->data_title[0] = '\0';
	m_pDiscData->data_artist[0] = '\0';
	m_pDiscData->data_extended[0] = '\0';
	
	unsigned long nDiscID = cddb_discid(m_nDevice);
	cout << hex << "disc id: " << nDiscID << endl;

	CheckCDDBMessage();

	cddb_stat_disc_data(m_nDevice, &m_Entry);
	if (m_Entry.entry_present) {
		EM::I()->Trace() << "entry present..." << endl;
		cddb_read_disc_data(m_nDevice, m_pDiscData.get());
		if (m_pDiscData.get() == NULL) {
			EM::I()->Trace() << "entry null." << endl;
		}
	}
	else {

		std::string strServer = System::GetPropList()->GetValue("CDDBServer");
		int nPort = System::GetPropList()->GetIntValue("CDDBServerPort");

		cddb_hello hello;
		strncpy(hello.hello_program, PACKAGE, 256);
		strncpy(hello.hello_version, VERSION, 256);

		cddb_server server;
		strcpy(server.server_name, strServer.c_str());
		server.server_port = nPort;

		cddb_host host;
		host.host_server = server;
		host.host_protocol = CDDB_MODE_CDDBP;
		
		m_nSocket = cddb_connect_server(host, NULL, hello);
		EM::I()->Noise() << "cddb_connect socket: " << m_nSocket << endl;
		CheckCDDBMessage();

		int nResult = cddb_query(m_nDevice, m_nSocket, CDDB_MODE_CDDBP,
								 &m_Query);
		EM::I()->Trace() << "cddb_query result: " << nResult << endl;
		CheckCDDBMessage();

		if (nResult == -1) {
			EM::I()->Error() << "cddb query error" << endl;
			cddb_generate_unknown_entry(m_nDevice, m_pDiscData.get());
			return;
		}
		
		EM::I()->Trace() << "query match: " << m_Query.query_match << endl;
		EM::I()->Trace() << "# of matches: " << m_Query.query_matches << endl;

		if (m_Query.query_match == QUERY_EXACT) {
			if (strlen(m_Query.query_list[0].list_artist) > 0) {
				EM::I()->Trace() << "Match for " << cddb_discid(m_nDevice) <<
					":  " << m_Query.query_list[0].list_artist << " / " <<
					m_Query.query_list[0].list_title << endl;
			}
			m_Entry.entry_genre = m_Query.query_list[0].list_genre;
			m_Entry.entry_id = m_Query.query_list[0].list_id;

			if (cddb_read(m_nDevice, m_nSocket, CDDB_MODE_CDDBP, m_Entry,
						  m_pDiscData.get()) < 0) {
				EM::I()->Error() << "cddb read error";
				cddb_generate_unknown_entry(m_nDevice, m_pDiscData.get());
				return;
			}
			cddb_quit(m_nSocket);
		}
		else if (m_Query.query_match == QUERY_INEXACT) {
			EM::I()->Trace() << "Inexact match for " <<
				cddb_discid(m_nDevice) << endl;

			ChoiceDialog* pDialog = ChoiceDialog::Instance();
			pDialog->SetText("Inexact CDDB match.  Please choose one:");
			pDialog->ClearChoices();
			m_Connection =
				pDialog->ChoiceDialogDone.connect(
					SigC::slot(this, &AudioCD::ChoiceDialogDone));
			
			for (int i = 0; i < m_Query.query_matches; i++) {
				if (strlen(m_Query.query_list[i].list_artist) < 1) {
					EM::I()->Trace() << i + 1 <<  ": " <<
						m_Query.query_list[i].list_title << endl;
					pDialog->AddChoice(m_Query.query_list[i].list_title);
				}
				else {
					std::string s = std::string(m_Query.query_list[i].list_artist) +
						" / " + m_Query.query_list[i].list_title;
					EM::I()->Trace() << i + 1 << ": " << s << endl;
					pDialog->AddChoice(s);
				}
			}

			pDialog->CenterOverDisplay();
			return;
			
		}
		else if (m_Query.query_match == QUERY_NOMATCH) {
			EM::I()->Trace() << "No match for " << hex <<
				cddb_discid(m_nDevice) << dec << endl;
			cddb_generate_unknown_entry(m_nDevice, m_pDiscData.get());
		}

		close(m_nSocket);
		cddb_write_data(m_nDevice, m_pDiscData.get());
	}
}

//------------------------------------------------------------------------------

void
AudioCD::CheckCDDBMessage() const
{
	if (strlen(cddb_message)) {
		EM::I()->Trace() << "cddb_message: " << cddb_message << endl;
	}
}

//------------------------------------------------------------------------------

void
AudioCD::Play(int nTrack)
{
	Assert(m_nDevice != s_nInvalidDevice, "cd device uninitialized");
	cd_play(m_nDevice, nTrack);
}

//------------------------------------------------------------------------------

void
AudioCD::Stop()
{
	EM::I()->Trace() << "stopping cd play" << endl;
	Assert(m_nDevice != s_nInvalidDevice, "cd device uninitialized");
	cd_stop(m_nDevice);
}

//------------------------------------------------------------------------------

void
AudioCD::Eject()
{
	Assert(m_nDevice != s_nInvalidDevice, "cd device uninitialized");
	cd_eject(m_nDevice);
}

//------------------------------------------------------------------------------

bool
AudioCD::CheckDiscInDrive() const
{
	bool bResult = true;
	
	if (m_nDevice == s_nInvalidDevice) {
		bResult = false;
	}
	else {
		struct disc_info info;
	
		cd_stat(m_nDevice, &info);
		if (!info.disc_present) {
//			cd_close(m_nDevice);
//			cd_stat(m_nDevice, &info);
//			if (!info.disc_present) {
//				EM::I()->Trace("No disc in drive");
				bResult = false;
//			}
		}
	}

	return bResult;
}

//------------------------------------------------------------------------------

void
AudioCD::ChoiceDialogDone(int n)
{
	EM::I()->Trace() << "dialog choice:  " << n << endl;
	m_Connection.disconnect();

	m_Entry.entry_genre = m_Query.query_list[n].list_genre;
	m_Entry.entry_id = m_Query.query_list[n].list_id;
	if (cddb_read(m_nDevice, m_nSocket, CDDB_MODE_CDDBP, m_Entry,
				  m_pDiscData.get()) < 0) {
		EM::I()->Error() << "CDDB read error" << endl;
		cddb_generate_unknown_entry(m_nDevice, m_pDiscData.get());
		return;
	}
	cddb_quit(m_nSocket);
	close(m_nSocket);
	cddb_write_data(m_nDevice, m_pDiscData.get());
	Update();
}

//------------------------------------------------------------------------------

bool
AudioCD::CheckCapability(std::string& strMessage)
{
	bool bResult = true;
	
	if (m_nDevice == s_nInvalidDevice) {
		std::string strDeviceName =	System::GetPropList()->GetValue("CDReader");
		strMessage = "Couldn't access cd reader device: " + strDeviceName;
		bResult = false;
	}
	
	return bResult;
}

//------------------------------------------------------------------------------

bool
AudioCD::MediaChanged()
{
	bool bResult = false;

#ifdef HAVE_LINUX_CDROM_H
	if (m_nDevice != s_nInvalidDevice) {
		if (CheckDiscInDrive()) {
			EM::I()->Trace() << "disc in drive..." << endl;
			int nResult = ioctl(m_nDevice, CDROM_MEDIA_CHANGED, CDSL_CURRENT);
			bResult = (nResult == 1);
		}
	}
#endif

	if (bResult == true) {
		Update();
	}
	
	return bResult;
}

//------------------------------------------------------------------------------

std::string
AudioCD::GetArtist() const
{
	std::string strResult;

	if (m_pDiscData.get() != NULL) {
		strResult = m_pDiscData->data_artist;
	}
	
	return strResult;
}

//------------------------------------------------------------------------------

std::string
AudioCD::GetTitle() const
{
	std::string strResult;

	if (m_pDiscData.get() != NULL) {
		strResult = m_pDiscData->data_title;
	}
	
	return strResult;
}

//------------------------------------------------------------------------------

bool
AudioCD::IsAudio() const
{
	bool bResult = false;
	
	cdrom_drive* pDrive;
	std::string strDevice =	System::GetPropList()->GetValue("CDReader");
	int nVerbose = CDDA_MESSAGE_PRINTIT;
	pDrive = cdda_identify(strDevice.c_str(), nVerbose, NULL);
	if (pDrive != NULL) {
		int nResult = cdda_open(pDrive);
		bResult = (nResult == 0);
		cdda_close(pDrive);
	}

	return bResult;
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
