// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// Controller.cpp
//
// $Id: Controller.cpp,v 1.53 2002/03/09 19:14:08 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "Controller.h"

#include <dlfcn.h>
#include <xb/AutoLockGtk.h>
#include <xb/ErrorManager.h>
#include <xb/FileUtil.h>
#include <xb/String.h>

#include "BurnManager.h"
#include "Info.h"
#include "Library.h"
#include "Mixer.h"
#include "Output.h"
#include "Player.h"
#include "PlayList.h"
#include "PlayListManager.h"
#include "PropList.h"
#include "RipManager.h"
#include "System.h"
#include "Util.h"
#include "Visualization.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

typedef	Info* (*info_maker_function)(void);
typedef	Output* (*output_maker_function)(void);
typedef	Player* (*player_maker_function)(void);

//------------------------------------------------------------------------------

Controller::Controller() :
	m_pAudioCD(new AudioCD()),
	m_pBurnManager(new BurnManager()),
	m_pMixer(new Mixer()),
	m_pPlayer(NULL),
	m_pRipManager(new RipManager()),
	m_pOutput(NULL),
	m_pVisualization(new Visualization()),
	m_bCDAvailable(false),
	m_bBurnAvailable(false),
	m_bRipAvailable(false)
{
	// audio cd
	// AssertNotNull(m_pAudioCD.get());

	// burn manager
	// AssertNotNull(m_pBurnManager.get());
	m_pBurnManager->StatusUpdate.connect(BurnStatusUpdate.slot());
	m_pBurnManager->Done.connect(BurnDone.slot());
	m_pBurnManager->Done.connect(SigC::slot(this, &Controller::BurnDoneImpl));
	
	// mixer
	// AssertNotNull(m_pMixer.get());

	// visualization
	// AssertNotNull(m_pVisualization.get());
	
	// output
	m_pOutput.reset(GetOutputPlugin("oss"));
	AssertNotNull(m_pOutput.get());
	m_pOutput->Init();

	// player
	LoadPlayerPlugins();
	
	// rip manager
	// AssertNotNull(m_pRipManager.get());
	m_pRipManager->StatusUpdate.connect(RipStatusUpdate.slot());
	m_pRipManager->Done.connect(RipDone.slot());
	m_pRipManager->Done.connect(SigC::slot(this, &Controller::RipDoneImpl));
}

//------------------------------------------------------------------------------

Controller::~Controller()
{
	Quit();
}

//------------------------------------------------------------------------------

void
Controller::Quit()
{
	PlayerList::const_iterator itr;
	for (itr = m_listPlayers.begin(); itr != m_listPlayers.end(); ++itr) {
		delete *itr;
	}
}

//------------------------------------------------------------------------------

bool
Controller::CheckCapability(std::string& strMessage)
{
	AssertNotNull(m_pAudioCD.get());
	AssertNotNull(m_pBurnManager.get());
	AssertNotNull(m_pMixer.get());
	AssertNotNull(m_pRipManager.get());
	
	bool bResult = true;
	std::string s;

	strMessage = "The following problems were detected:\n";
	
	if (!m_pAudioCD->CheckCapability(s)) {
		SetCDAvailable(false);
		strMessage += s + "\n";
		bResult = false;
	}
	else {
		SetCDAvailable(true);
	}

	if (!m_pBurnManager->CheckCapability(s)) {
		strMessage += s + "\n";
		bResult = false;
		m_bBurnAvailable = false;
	}
	else {
		m_bBurnAvailable = true;
	}

	if (!m_pMixer->CheckCapability(s)) {
		strMessage += s + "\n";
		bResult = false;
	}

	if (!m_pRipManager->CheckCapability(s)) {
		strMessage += s + "\n";
		bResult = false;
		m_bRipAvailable = false;
	}
	else {
		m_bRipAvailable = true;
	}

	return bResult;
}

//------------------------------------------------------------------------------

int
Controller::GetVolume()
{
	AssertNotNull(m_pMixer.get());
	return m_pMixer->GetVolume();
}

//------------------------------------------------------------------------------

void
Controller::SetVolume(int nVal)
{
	AssertNotNull(m_pMixer.get());
	m_pMixer->SetVolume(nVal);
}

//------------------------------------------------------------------------------

void
Controller::PlayerStop(bool bSendSignal)
{
	if (m_pPlayer != NULL) {
		m_pPlayer->Stop(bSendSignal);

		if (m_pVisualization.get() != NULL) {
			m_pVisualization->Reset();
		}
	}
}

//------------------------------------------------------------------------------

void
Controller::PlayerPause(bool bPause)
{
	AssertNotNull(m_pPlayer);
	m_pPlayer->Pause(bPause);
}

//------------------------------------------------------------------------------

void
Controller::PlayerPlay(const xb::Filename& strFilename)
{
	SetPlayerPluginForFile(strFilename);

	AssertNotNull(m_pPlayer);
	m_pPlayer->Init(m_pOutput.get(), m_pVisualization.get());
	m_pPlayer->PlayComplete.connect(PlayerComplete.slot());
	
	m_pPlayer->Play(strFilename, 0);
}

//------------------------------------------------------------------------------

bool
Controller::IsPlaying() const
{
	bool bResult = false;

	if (m_pPlayer) {
		bResult = m_pPlayer->IsPlaying();
	}

	return bResult;
}

//------------------------------------------------------------------------------

bool
Controller::IsPlayComplete() const
{
	AssertNotNull(m_pPlayer);
	return m_pPlayer->IsPlayComplete();
}

//------------------------------------------------------------------------------

int
Controller::GetPlayerElapsedTime() const
{
	AssertNotNull(m_pPlayer);
	return m_pPlayer->GetTime();
}

//------------------------------------------------------------------------------

void
Controller::PlayerSeek(float fPercentage)
{
	AssertNotNull(m_pPlayer);
	m_pPlayer->SeekPercentage(fPercentage);
}

//------------------------------------------------------------------------------

void
Controller::CreateNewPlayList()
{
	System::GetPlayListManager()->NewPlayList();
}

//------------------------------------------------------------------------------

PlayList*
Controller::GetPlayList(const std::string& strName)
{
	return System::GetPlayListManager()->GetPlayList(strName);
}

//------------------------------------------------------------------------------

void
Controller::RenamePlayList(const std::string& strOriginal,
						   const std::string& strNew)
{
	System::GetPlayListManager()->Rename(strOriginal, strNew);
}

//------------------------------------------------------------------------------

void
Controller::UpdatePlayListSystem()
{
	System::GetPlayListManager()->UpdateDirectoryList();
}

//------------------------------------------------------------------------------

void
Controller::DeletePlayList(const std::string& strName)
{
	System::GetPlayListManager()->DeletePlayList(strName);
	UpdatePlayListSystem();
}

//------------------------------------------------------------------------------

void
Controller::DeletePlayListSongs(const std::string& strName,
								const SongList& listSongs)
{
	PlayList* pPlayList = GetPlayList(strName);
	AssertNotNull(pPlayList);
	
	SongList::const_iterator itr;
	for (itr = listSongs.begin(); itr != listSongs.end(); ++itr) {
		pPlayList->Remove(*itr);
	}
	pPlayList->Save();
}

//------------------------------------------------------------------------------

void
Controller::DeleteLibrarySongs(const SongList& listSongs)
{
	SongList::const_iterator itr;
	for (itr = listSongs.begin(); itr != listSongs.end(); ++itr) {
		System::GetLibrary()->Remove(*itr);
	}

}

//------------------------------------------------------------------------------

AudioCD*
Controller::GetAudioCD()
{
	return m_pAudioCD.get();
}

//------------------------------------------------------------------------------

void
Controller::CDEject()
{
	AssertNotNull(m_pAudioCD.get());
	m_pAudioCD->Eject();
}

//------------------------------------------------------------------------------

void
Controller::SetCDAvailable(bool bIn)
{
	if (m_bCDAvailable != bIn) {
		m_bCDAvailable = bIn;
		CDAvailableChange(m_bCDAvailable);
	}
}

//------------------------------------------------------------------------------

void
Controller::CDPlay(int nTrack)
{
	AssertNotNull(m_pAudioCD.get());
	EM::I()->Trace() << "Playing audio cd, track " << nTrack << endl;
	m_pAudioCD->Play(nTrack);
}

//------------------------------------------------------------------------------

void
Controller::CDStop()
{
	AssertNotNull(m_pAudioCD.get());
	EM::I()->Trace() << "Stopping audio cd." << endl;;
	m_pAudioCD->Stop();
}

//------------------------------------------------------------------------------

bool
Controller::CDMediaChanged()
{
	AssertNotNull(m_pAudioCD.get());
	return m_pAudioCD->MediaChanged();
}

//------------------------------------------------------------------------------

std::string
Controller::CDGetArtist() const
{
	AssertNotNull(m_pAudioCD.get());
	return m_pAudioCD->GetArtist();
}

//------------------------------------------------------------------------------

std::string
Controller::CDGetTitle() const
{
	AssertNotNull(m_pAudioCD.get());
	return m_pAudioCD->GetTitle();
}

//------------------------------------------------------------------------------

void
Controller::Burn(const SongList& listSongs, bool bNoAction)
{
	AssertNotNull(m_pBurnManager.get());
	Assert(m_bBurnAvailable, "burn unavailable");
	SetCDAvailable(false);
	m_bRipAvailable = false;
	m_bBurnAvailable = false;
	m_pBurnManager->Burn(listSongs, bNoAction);
}

//------------------------------------------------------------------------------

void
Controller::BurnDoneImpl()
{
//	xb::AutoLockGtk lock("BurnDoneImpl");
//	
	SetCDAvailable(true);
	m_bRipAvailable = true;
	m_bBurnAvailable = true;
}

//------------------------------------------------------------------------------

void
Controller::BurnCancel()
{
	AssertNotNull(m_pBurnManager.get());
	m_pBurnManager->Cancel();
	BurnDoneImpl();
}

//------------------------------------------------------------------------------

void
Controller::Rip(const SongList& listSongs, bool bNoAction)
{
	AssertNotNull(m_pRipManager.get());
	Assert(m_bCDAvailable, "cd unavailable");
	Assert(m_bRipAvailable, "rip unavailable");
	m_bRipAvailable = false;
	m_bBurnAvailable = false;

	std::string strDevice = System::GetPropList()->GetValue("RipDevice");
	
	m_pRipManager->Rip(strDevice, listSongs, bNoAction);
}

//------------------------------------------------------------------------------

void
Controller::RipDoneImpl()
{
	m_bRipAvailable = true;
	m_bBurnAvailable = true;
}

//------------------------------------------------------------------------------

void
Controller::RipCancel()
{
	AssertNotNull(m_pRipManager.get());
	m_pRipManager->Cancel();
	RipDoneImpl();
}

//------------------------------------------------------------------------------

void
Controller::UpdateVisualization(int nTime)
{
	AssertNotNull(m_pVisualization.get());

	m_pVisualization->Update(nTime);
}

//------------------------------------------------------------------------------

const gint16*
Controller::GetVisualizationData() const
{
	AssertNotNull(m_pVisualization.get());

	return m_pVisualization->GetData();
}

//------------------------------------------------------------------------------

int
Controller::GetVisualizationChannels() const
{
	AssertNotNull(m_pVisualization.get());

	return m_pVisualization->GetChannels();
}

//------------------------------------------------------------------------------
// caller responsible for deleting result when done with it...

Player*
Controller::GetPlayerPlugin(const std::string& strName)
{
	Player* pResult = NULL;
	
	std::string strPath = PLUGIN_DIR "/input/lib";
	strPath += strName + ".so";

	EM::I()->Trace() << "opening plugin: " << strPath << endl;
	
	void* pLib = dlopen(strPath.c_str(), RTLD_NOW);
	if (pLib == NULL) {
		EM::I()->Error() << "couldn't open " << strPath << ": " <<
			dlerror() << endl;
	}
	else {
		dlerror();

		player_maker_function mkr = (player_maker_function)dlsym(pLib, "maker");
		const char* pError = dlerror();
		if (pError) {
			EM::I()->Error() << "couldn't find maker in " << strPath << endl;
		}
		else {
			pResult = mkr();
		}
	}

	return pResult;
}

//------------------------------------------------------------------------------
// caller responsible for deleting result

Player*
Controller::GetPlayerPluginForFile(const xb::Filename& strFilename)
{
	Player* pResult = NULL;
	
	std::string strPath = PLUGIN_DIR "/input/";

	xb::FileUtil::DirectoryList listDir;
	xb::FileUtil::GetDirectoryList(strPath, listDir);
	xb::FileUtil::DirectoryListConstItr itr;
	for (itr = listDir.begin(); itr != listDir.end(); ++itr) {

		char* ext = strrchr(itr->c_str(), '.');
		if (ext && !strncasecmp(ext, ".so", 3)) {
		
			std::string strFull(strPath + *itr);
			EM::I()->Trace() << "Controller::GetPlayerPlugin() -- " <<
				strFull << endl;

			void* pLib = dlopen(strFull.c_str(), RTLD_NOW);
			if (pLib == NULL) {
				EM::I()->Error() << "couldn't open " << strFull << ": " <<
					dlerror() << endl;
			}
			else {
				dlerror();

				player_maker_function mkr =
					(player_maker_function)dlsym(pLib, "maker");
				const char* pError = dlerror();
				if (pError) {
					EM::I()->Error() << "couldn't find maker in " << strFull <<
						endl;
				}
				else {
					Player* pPlayer = mkr();
					if (pPlayer && pPlayer->CanPlayFile(strFilename)) {
						pResult = pPlayer;
					}
					else {
						delete pPlayer;
					}
				}
			}
		}
	}
	
	return pResult;
}

//------------------------------------------------------------------------------
// caller responsible for deleting result

Output*
Controller::GetOutputPlugin(const std::string& strName)
{
	Output* pResult = NULL;
	
	std::string strPath = PLUGIN_DIR "/output/lib" + strName + ".so";
	void* pOutLib = dlopen(strPath.c_str(), RTLD_NOW);
	if (pOutLib == NULL) {
		EM::I()->Error() << "couldn't open " << strPath << ": " << dlerror() <<
			endl;
	}
	else {
		dlerror();

		output_maker_function out_mkr =
			(output_maker_function)dlsym(pOutLib, "maker");
		const char* pError = dlerror();
		if (pError) {
			EM::I()->Error() << "couldn't find maker in " << strPath << endl;
		}
		else {
			pResult = out_mkr();
		}
	}

	return pResult;
}

//------------------------------------------------------------------------------

void
Controller::LoadPlayerPlugins()
{
	std::string strPath = PLUGIN_DIR "/input/";

	xb::FileUtil::DirectoryList listDir;
	xb::FileUtil::GetDirectoryList(strPath, listDir);
	xb::FileUtil::DirectoryListConstItr itr;
	for (itr = listDir.begin(); itr != listDir.end(); ++itr) {

		char* ext = strrchr(itr->c_str(), '.');
		if (ext && !strncasecmp(ext, ".so", 3)) {
		
			std::string strFull(strPath + *itr);
			EM::I()->Trace() << "Controller::LoadPlayerPlugins() -- " <<
				strFull << endl;

			void* pLib = dlopen(strFull.c_str(), RTLD_NOW);
			if (pLib == NULL) {
				EM::I()->Error() << "couldn't open " << strFull << ": " <<
					dlerror() << endl;
			}
			else {
				dlerror();

				player_maker_function mkr =
					(player_maker_function)dlsym(pLib, "maker");
				const char* pError = dlerror();
				if (pError) {
					EM::I()->Error() << "couldn't find maker in " << strFull <<
						endl;
				}
				else {
					Player* pPlayer = mkr();

					m_listPlayers.push_back(pPlayer);
					
					/*
					if (pPlayer && pPlayer->CanPlayFile(strFilename)) {
						pResult = pPlayer;
					}
					else {
						delete pPlayer;
					}
					*/
				}
			}
		}
	}
}

//------------------------------------------------------------------------------

void
Controller::SetPlayerPluginForFile(const xb::Filename& strFilename)
{
	m_pPlayer = NULL;
	
	PlayerList::const_iterator itr;
	for (itr = m_listPlayers.begin();
		 itr != m_listPlayers.end() && m_pPlayer == NULL;
		 ++itr) {
		
		if ((*itr)->CanPlayFile(strFilename)) {
			m_pPlayer = *itr;
		}
			
	}
}

//------------------------------------------------------------------------------
// caller responsible for deleting result

Info*
Controller::GetInfoPluginForFile(const xb::Filename& strFilename)
{
	Info* pResult = NULL;
	
	std::string strPath = PLUGIN_DIR "/info/";

	xb::FileUtil::DirectoryList listDir;
	xb::FileUtil::GetDirectoryList(strPath, listDir);
	xb::FileUtil::DirectoryListConstItr itr;
	for (itr = listDir.begin(); itr != listDir.end(); ++itr) {

		char* ext = strrchr(itr->c_str(), '.');
		if (ext && !strncasecmp(ext, ".so", 3)) {
		
			std::string strFull(strPath + *itr);
			// EM::I()->Trace("Controller::GetInfoPluginForFile() -- " + strFull);

			void* pLib = dlopen(strFull.c_str(), RTLD_NOW);
			if (pLib == NULL) {
				EM::I()->Error() << "couldn't open " << strFull << ": " <<
					dlerror() << endl;
			}
			else {
				dlerror();

				info_maker_function mkr =
					(info_maker_function)dlsym(pLib, "maker");
				const char* pError = dlerror();
				if (pError) {
					EM::I()->Error() << "couldn't find maker in " << strFull <<
						endl;
				}
				else {
					Info* pInfo = mkr();
					if (pInfo && pInfo->CanReadFile(strFilename)) {
						pResult = pInfo;
					}
					else {
						delete pInfo;
					}
				}
			}
		}
	}

	if (pResult != NULL) {
		pResult->Init(strFilename);
	}
	else {
		EM::I()->Warn() << "couldn't find info plugin for " << strFilename <<
			endl;
	}
	
	return pResult;
}

//------------------------------------------------------------------------------

void
Controller::GetFileOutputPluginList(PluginList& plugins)
{
	plugins.clear();

	std::string strPath = PLUGIN_DIR "/output/";

	xb::FileUtil::DirectoryList listDir;
	xb::FileUtil::GetDirectoryList(strPath, listDir);
	xb::FileUtil::DirectoryListConstItr itr;
	for (itr = listDir.begin(); itr != listDir.end(); ++itr) {

		char* ext = strrchr(itr->c_str(), '.');
		if (ext && !strncasecmp(ext, ".so", 3)) {
		
			std::string strFull(strPath + *itr);

			EM::I()->Trace() << "plugin: " << strFull << endl;
			
			void* pLib = dlopen(strFull.c_str(), RTLD_NOW);
			if (pLib == NULL) {
				EM::I()->Error() << "couldn't open " << strFull << ": " <<
					dlerror() << endl;
			}
			else {
				dlerror();

				output_maker_function mkr =
					(output_maker_function)dlsym(pLib, "maker");
				const char* pError = dlerror();
				if (pError) {
					EM::I()->Error() << "couldn't find maker in " << strFull <<
						endl;
				}
				else {
					Output* pOutput = mkr();
					if (pOutput != NULL && pOutput->IsFileOutput()) {
						PluginInfo info;
						info.m_strLibName = pOutput->GetLibName();
						info.m_strDescription = pOutput->GetDescription();

						plugins.push_back(info);
					}
				}
			}
		}
	}
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------







