// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// Controller.h
//
// $Id: Controller.h,v 1.34 2002/02/18 08:55:58 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#ifndef __Controller__
#define __Controller__

#include <glib.h>
#include <list>
#include <memory>
#include <sigc++/object.h>
#include <sigc++/signal_system.h>
#include <string>
#include <vector>
#include <xb/Filename.h>

#include "AudioCD.h"
#include "PlayList.h"
#include "SongList.h"

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

class BurnManager;
class Info;
class Mixer;
class Output;
class Player;
class RipManager;
class Visualization;

class Controller : public SigC::Object
{
	// Constants
	// Types
	private:
		typedef std::vector<Player*> PlayerList;

	public:
		struct PluginInfo {
				std::string m_strLibName;
				std::string m_strDescription;
		};

		typedef std::list<PluginInfo> PluginList;
		
	// Variables
	private:
		auto_ptr<AudioCD> m_pAudioCD;
		auto_ptr<BurnManager> m_pBurnManager;
		auto_ptr<Mixer> m_pMixer;
	// auto_ptr<Player> m_pPlayer;
		Player* m_pPlayer;
		auto_ptr<RipManager> m_pRipManager;
		auto_ptr<Output> m_pOutput;
		auto_ptr<Visualization> m_pVisualization;
		
		bool m_bCDAvailable;
		bool m_bBurnAvailable;
		bool m_bRipAvailable;

		PlayerList m_listPlayers;
		
	// Methods
	public:
		Controller();
		~Controller();

		void Quit();
		
		bool CheckCapability(std::string& strMessage);
		
		// mixer methods
		int GetVolume();
		void SetVolume(int nVal);

		// mp3 player methods
		SigC::Signal0<void> PlayerComplete;
		void PlayerStop(bool bSendSignal = true);
		void PlayerPause(bool bPause);
		void PlayerPlay(const xb::Filename& strFilename);
		bool IsPlaying() const;
		bool IsPlayComplete() const;
		int GetPlayerElapsedTime() const;
		void PlayerSeek(float fPercentage);

		// playlists
		void CreateNewPlayList();
		PlayList* GetPlayList(const std::string& strName);
		void RenamePlayList(const std::string& strOriginal,
							const std::string& strNew);
		void UpdatePlayListSystem();
		void DeletePlayList(const std::string& strName);
		void DeletePlayListSongs(const std::string& strName,
								 const SongList& listSongs);

		// library methods
		void DeleteLibrarySongs(const SongList& listSongs);

		// cd methods
		AudioCD* GetAudioCD();
		void CDEject();
		bool CDAvailable() const { return m_bCDAvailable; }
		SigC::Signal1<void, bool> CDAvailableChange;
		void CDPlay(int nTrack);
		void CDStop();
		bool CDMediaChanged();
		std::string CDGetArtist() const;
		std::string CDGetTitle() const;

		// burn methods
		void Burn(const SongList& listSongs, bool bNoAction);
		SigC::Signal1<void, const std::string&> BurnStatusUpdate;
		SigC::Signal0<void> BurnDone;
		bool BurnAvailable() const { return m_bBurnAvailable; }
		void BurnCancel();
		
		// rip methods
		void Rip(const SongList& listSongs, bool bNoAction);
		SigC::Signal1<void, const std::string&> RipStatusUpdate;
		SigC::Signal0<void> RipDone;
		bool RipAvailable() const { return m_bRipAvailable; }
		void RipCancel();

		// visualization methods
		void UpdateVisualization(int nTime);
		const gint16* GetVisualizationData() const;
		int GetVisualizationChannels() const;

		// plugin methods
		static Player* GetPlayerPlugin(const std::string& strName);
		static Player* GetPlayerPluginForFile(const xb::Filename& strFilename);
		static Output* GetOutputPlugin(const std::string& strName);
		static Info* GetInfoPluginForFile(const xb::Filename& strFilename);
		
		static void GetFileOutputPluginList(PluginList& plugins);
		
		
	private:
		void SetCDAvailable(bool bIn);
		void RipDoneImpl();
		void BurnDoneImpl();
		void LoadPlayerPlugins();
		void SetPlayerPluginForFile(const xb::Filename& strFilename);
};

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

#endif

//------------------------------------------------------------------------------
