// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// Library.cpp
//
// $Id: Library.cpp,v 1.25 2002/02/26 19:30:27 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "Library.h"

#include <dirent.h>
#include <errno.h>
#include <fstream>
#include <stdio.h>
#include <xb/ErrorManager.h>
#include <xb/FileUtil.h>

#include "Controller.h"
#include "Info.h"
#include "PropList.h"
#include "Source.h"
#include "System.h"
#include "Util.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

Library::Library(const std::string& strRoot) :
	Source(),
	m_strRoot(strRoot),
	m_bDirty(false),
	m_nIDCount(0)
{
	xb::FileUtil::AssertDirectoryExists(m_strRoot);
	m_strCacheFilename = System::GetPropList()->GetValue("LibraryCache");

	if (CacheUpdateNeeded()) {
		Update();
	}
	
	if (ReadCache() == false) {
		Update();
	}
}

//------------------------------------------------------------------------------

Library::~Library()
{
	m_SongList.clear();
}

//------------------------------------------------------------------------------

void
Library::Update()
{
	m_SongList.clear();
	
	xb::FileUtil::DirectoryList listArtists;
	bool bResult = xb::FileUtil::GetDirectoryList(m_strRoot, listArtists);
	if (!bResult) {
		throw(xb::file_error("directory list error."));
	}

	xb::FileUtil::DirectoryListConstItr itrArtist;
	for (itrArtist = listArtists.begin();
		 itrArtist != listArtists.end();
		 ++itrArtist) {
		
		EM::I()->Noise() << *itrArtist << endl;

		std::string strArtistPath = m_strRoot;
		strArtistPath.append("/");
		strArtistPath.append(*itrArtist);
		xb::FileUtil::DirectoryList listAlbums;
		bResult = xb::FileUtil::GetDirectoryList(strArtistPath, listAlbums);
		if (!bResult) {
			throw(xb::file_error("directory list error."));
		}

		xb::FileUtil::DirectoryListConstItr itrAlbum;
		for (itrAlbum = listAlbums.begin(); itrAlbum != listAlbums.end();
			 ++itrAlbum) {

			std::string strAlbumPath = strArtistPath;
			strAlbumPath.append("/");
			strAlbumPath.append(*itrAlbum);
			xb::FileUtil::DirectoryList listSongs;
			bResult = xb::FileUtil::GetDirectoryList(strAlbumPath, listSongs);
			if (!bResult) {
				throw(xb::file_error("directory list error."));
			}

			
			xb::FileUtil::DirectoryListConstItr itrSong;
			for (itrSong = listSongs.begin(); itrSong != listSongs.end();
				 ++itrSong) {

				std::string strTitle = *itrSong;
				std::string strFilename = strAlbumPath + "/" + strTitle;
				AddSong(strFilename, strTitle, *itrArtist, *itrAlbum);
			}
		}
	}

	WriteCache();
	SetCacheUpdate(false);
}

//------------------------------------------------------------------------------

void
Library::AddSong(const std::string& strFilename,
				 const std::string& strTitleIn,
				 const std::string& strArtist,
				 const std::string& strAlbum)
{
	std::string strTitle = strTitleIn;
	// PATRICK: check to make sure .mp3 in title
	// remove .mp3 from title
	int nLength = strTitle.length();
	strTitle.erase(nLength - 4, nLength);
	
	Song song;
	song.SetID(m_nIDCount++);
	song.SetArtist(strArtist);
	song.SetAlbum(strAlbum);
	song.SetTitle(strTitle);
	song.SetFilename(strFilename);

	Info* pInfo = System::GetController()->GetInfoPluginForFile(strFilename);

	if (pInfo == NULL) {
		EM::I()->Warn() << "no info plugin for " <<  strFilename << endl;
		return;
	}
	
	bool bConvert = false;
	if (pInfo->GetArtist() != song.GetArtist()) {
		EM::I()->Trace() << "artists don't match: mp3info: " <<
			pInfo->GetArtist() << ", lib: " << song.GetArtist() << endl;
		if (pInfo->GetArtist().length() > 0) {
			song.SetArtist(pInfo->GetArtist());
		}
		else {
			bConvert = true;
		}
	}

	if (pInfo->GetAlbum() != song.GetAlbum()) {
		EM::I()->Trace() << "albums don't match: mp3info: " <<
			pInfo->GetAlbum() << ", lib: " << song.GetAlbum() << endl;
		if (pInfo->GetAlbum().length() > 0) {
			song.SetAlbum(pInfo->GetAlbum());
		}
		else {
			bConvert = true;
		}
	}
				
	if (pInfo->GetTitle() != song.GetTitle()) {
		EM::I()->Trace() << "titles don't match: mp3info: " << 
			pInfo->GetTitle() << ", lib: " << song.GetTitle() << endl;
		if (pInfo->GetTitle().length() > 0) {
			song.SetTitle(pInfo->GetTitle());
		}
		else {
			bConvert = true;
		}
	}

	song.SetSeconds(pInfo->GetSeconds());
	song.SetTime(Util::GetTimeStr(song.GetSeconds()));

	if (bConvert) {
		// PATRICK: fix
		// pInfo->WriteTag(song);
	}
				
	m_SongList.push_back(song);
}

//------------------------------------------------------------------------------

void
Library::Dump() const
{
	EM::I()->Trace() << "Root: " << m_strRoot << endl;

	m_SongList.Dump();
}

//------------------------------------------------------------------------------

xb::Filename
Library::GetFilename(const Song& song) const
{
	std::string strFilename;

	strFilename = m_strRoot + "/" + song.GetArtist() + "/" +
		song.GetAlbum() + "/" +
		song.GetTitle() + ".mp3";

	return strFilename;
}

//------------------------------------------------------------------------------

void
Library::Remove(const Song& song)
{
	bool bFound = false;
	SongList::iterator itr;
	for (itr = m_SongList.begin(); itr != m_SongList.end() && !bFound; ++itr) {
		if (*itr == song) {
			EM::I()->Trace() << "removing " << song.GetTitle() <<
				" from library." << endl;
			m_SongList.erase(itr);
			xb::FileUtil::Delete(song.GetFilename());
			bFound = true;
		}
	}
	
	if (!bFound) {
		EM::I()->Warn() << "couldn't find " << song.GetTitle() <<
			" in library." << endl;
	}
	else {
		WriteCache();
	}
}

//------------------------------------------------------------------------------

std::string
Library::Import(const Song& song, const std::string& strInFilename, bool bLink)
{
	Assert(m_strRoot.length() > 0, "root is empty");

	std::string strExtension = xb::FileUtil::GetExtension(strInFilename);
	
	std::string strFilename(m_strRoot);
	xb::FileUtil::AssertDirectoryExists(strFilename);
	std::string strArtist = xb::FileUtil::CleanFilename(song.GetArtist());
	if (strArtist.length() == 0) {
		strArtist = "(unknown)";
	}
	strFilename += "/" + strArtist;
	xb::FileUtil::AssertDirectoryExists(strFilename);
	std::string strAlbum = xb::FileUtil::CleanFilename(song.GetAlbum());
	if (strAlbum.length() == 0) {
		strAlbum = "(unknown)";
	}
	strFilename += "/" + strAlbum; 
	xb::FileUtil::AssertDirectoryExists(strFilename);
	std::string strTitle = xb::FileUtil::CleanFilename(song.GetTitle());
	if (strTitle.length() == 0) {
		strTitle = "(unknown)";
	}
	strFilename += "/" + strTitle;

	strFilename = xb::FileUtil::GetAvailableFile(strFilename, strExtension);
	
	EM::I()->Trace() << "imported filename: " << strFilename << endl;

	if (bLink) {
		xb::FileUtil::Link(strInFilename, strFilename);
	}
	else {
		xb::FileUtil::Rename(strInFilename, strFilename);
		Info* pInfo = System::GetController()->GetInfoPluginForFile(strFilename);

		if (pInfo == NULL) {
			EM::I()->Warn() << "no info plugin for " << strFilename << endl;
			return "";
		}
		
		pInfo->SetArtist(strArtist);
		pInfo->SetAlbum(strAlbum);
		pInfo->SetTitle(strTitle);
		pInfo->SetTrack(song.GetTrack());
		pInfo->Save();
	}

	AddSong(strFilename, strTitle + strExtension, strArtist, strAlbum);
	
	// SetCacheUpdate(true);
	WriteCache();

	return strFilename;
}

//------------------------------------------------------------------------------

std::string
Library::Import(const std::string& strFilename, bool bLink)
{
	Assert(m_strRoot.length() > 0, "root is empty");
	std::string strResult;
	
	if (xb::FileUtil::IsRegular(strFilename)) {
		Song song(strFilename);
		strResult = Import(song, strFilename, bLink);
	}
	else if (xb::FileUtil::IsDirectory(strFilename)) {
		ImportDirectory(strFilename, bLink);
	}
	else if (xb::FileUtil::IsLink(strFilename)) {
		xb::Filename strLink = xb::FileUtil::GetLink(strFilename);
		Song song(strLink);
		// only link to links (don't move them)
		strResult = Import(song, strFilename, true);
	}

	return strResult;
}

//------------------------------------------------------------------------------

void
Library::ImportDirectory(const std::string& strFilename, bool bLink)
{
	if (xb::FileUtil::IsDirectory(strFilename)) {

		xb::FileUtil::DirectoryList listDir;
		bool bResult = xb::FileUtil::GetDirectoryList(strFilename, listDir);
		if (!bResult) {
			throw(xb::file_error("directory list error."));
		}	

		xb::FileUtil::DirectoryListConstItr itr;
		for (itr = listDir.begin(); itr != listDir.end(); ++itr) {
			xb::Filename strFull(strFilename);
			strFull += "/" + *itr;
			EM::I()->Trace() << "Import Directory, filename = " << strFull <<
				endl;
			if (!xb::FileUtil::IsDirectory(strFull)) {
				Song song(strFull);
				if (song.GetSeconds() > 0) {
					EM::I()->Trace() << "found mp3: " << strFull << " (" <<
						song.GetSeconds() << "s)" << endl;
					Import(strFull, bLink);
				}
			}
		}
	}
}

//------------------------------------------------------------------------------

const SongList&
Library::GetSongList()
{
	if (CacheUpdateNeeded()) {
		Update();
	}

	return Source::GetSongList();
}

//------------------------------------------------------------------------------

void
Library::WriteCache() const
{
	EM::I()->Trace() << "writing cache: " << m_strCacheFilename << endl;
	
	ofstream os(m_strCacheFilename.c_str());
	if (!os.is_open()) {
		throw(xb::file_error("couldn't open cache file:  " + m_strCacheFilename));
	}
	
	SongList::const_iterator itr;
	for (itr = m_SongList.begin(); itr != m_SongList.end(); ++itr) {
		os << *itr;
	}
	os.close();
}

//------------------------------------------------------------------------------

bool
Library::ReadCache()
{
	if (!xb::FileUtil::FileExists(m_strCacheFilename)) {
		return false;
	}

	EM::I()->Trace() << "found cache file, loading it." << endl;
	
	m_SongList.clear();
	ifstream is(m_strCacheFilename.c_str());
	while (!is.eof()) {
		Song song;
		is >> song;

		// EM::I()->Trace("cache song: ");
		// song.Dump();

		if (song.GetFilename().length() > 0) {
			m_SongList.push_back(song);
		}
	}

	is.close();
	
	SetCacheUpdate(false);
	return true;
}

//------------------------------------------------------------------------------

void
Library::SetCacheUpdate(bool bIn)
{
	m_bDirty = bIn;

	if (bIn) {
		xb::FileUtil::CreateFile(
			System::GetPropList()->GetValue("CacheUpdateFile"));
	}
	else {
		xb::FileUtil::Delete(
			System::GetPropList()->GetValue("CacheUpdateFile"));
	}
}

//------------------------------------------------------------------------------

void
Library::ResetTimes()
{
	SongList::iterator itr;
	for (itr = m_SongList.begin(); itr != m_SongList.end(); ++itr) {
		// PATRICK: fix
		// MP3Info2::Convert(itr->GetFilename());
	}
}

//------------------------------------------------------------------------------

void
Library::ClearCache()
{
	Update();
}

//------------------------------------------------------------------------------

bool
Library::CacheUpdateNeeded() const
{
	if (m_bDirty) {
		return true;
	}

	if (xb::FileUtil::FileExists(
		System::GetPropList()->GetValue("CacheUpdateFile"))) {
		
		return true;
		
	}

	return false;
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
