// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// MP3Info.cpp
//
// $Id: MP3Info.cpp,v 1.9 2002/02/26 19:30:27 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "MP3Info.h"

#include <sys/stat.h>
#include <stdlib.h>
#include <xb/ErrorManager.h>

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

const int MP3Info::NUM_SAMPLES;
const int MP3Info::FRAME_HEADER_SIZE;
const int MP3Info::MIN_FRAME_SIZE;
const int MP3Info::MIN_CONSEC_GOOD_FRAMES;

const int MP3Info::frame_size_index[3] = {24000, 72000, 72000};
const int MP3Info::bitrate[2][3][14] = { 
	{ /* MPEG 2.0 */
		{32,48,56,64,80,96,112,128,144,160,176,192,224,256},  /* layer 1 */
		{8,16,24,32,40,48,56,64,80,96,112,128,144,160},       /* layer 2 */
		{8,16,24,32,40,48,56,64,80,96,112,128,144,160}        /* layer 3 */
	},
	
	{ /* MPEG 1.0 */
		{32,64,96,128,160,192,224,256,288,320,352,384,416,448}, /* layer 1 */
		{32,48,56,64,80,96,112,128,160,192,224,256,320,384},    /* layer 2 */
		{32,40,48,56,64,80,96,112,128,160,192,224,256,320}      /* layer 3 */
	}
};

const int MP3Info::frequencies[3][4] = {
	{22050,24000,16000,50000},  /* MPEG 2.0 */
	{44100,48000,32000,50000},  /* MPEG 1.0 */
	{11025,12000,8000,50000}    /* MPEG 2.5 */
};

//------------------------------------------------------------------------------


MP3Info::MP3Info(const xb::Filename& strFilename) :
	m_strFilename(strFilename)
{
	bool bResult = LoadMP3Info();
	if (!bResult) {
		EM::I()->Error() << "error getting mp3info for " << m_strFilename <<
			endl;
	}
}

//------------------------------------------------------------------------------


MP3Info::~MP3Info()
{

}

//------------------------------------------------------------------------------

bool
MP3Info::LoadMP3Info(bool bFullScanVBR)
{
	bool bResult = true;
	off_t sample_pos, data_start;
	int nLastRate = 0;
	int nCounter = 0;
	int nBitRate = 0;

	m_pFile = fopen(m_strFilename.c_str(), "r");
	if (m_pFile == NULL) {
		return false;
	}
	
	struct stat filestat;
	stat(m_strFilename.c_str(), &filestat);
	m_DataSize = filestat.st_size;

	if (LoadID3()) {
		if (GetFirstHeader(0L)) {
			data_start = ftell(m_pFile);
			nLastRate = 15 - m_Header.m_nBitRate;
			while ((nCounter < NUM_SAMPLES) && nLastRate) {
				sample_pos = (nCounter * (m_DataSize / NUM_SAMPLES + 1)) + data_start;
				if (GetFirstHeader(sample_pos)) {
					nBitRate = 15 - m_Header.m_nBitRate;
				}
				else {
					nBitRate = -1;
				}
				
				if (nBitRate != nLastRate) {
					m_nVBR = 1;
					
					if (bFullScanVBR) {
						nCounter = NUM_SAMPLES;
						// nScanType = SCAN_FULL;
					}
				}
				nLastRate = nBitRate;
				nCounter++;
			}
				
			m_nFrames = (m_DataSize - data_start) / FrameLength(&m_Header);
			m_nSeconds = (int)((float)(FrameLength(&m_Header) * m_nFrames) /
							   (float)(HeaderBitRate(&m_Header) * 125) + 0.5);
			m_fVBRAverage = (float)HeaderBitRate(&m_Header);
		}
	}
	else {
		bResult = false;
	}

	fclose(m_pFile);
	
	return bResult;
}

//------------------------------------------------------------------------------

bool
MP3Info::LoadID3()
{
	bool bResult = false;

	char acBuf[4];

	if (m_DataSize >= 128) {
		if (fseek(m_pFile, -128, SEEK_END)) {
			EM::I()->Error() << "ERROR: Couldn't read last 128 bytes of " <<
				m_strFilename << endl;
			bResult = false;
		}
		else {
			bResult = true;
			
			fread(acBuf, 1, 3, m_pFile);
			acBuf[3] = '\0';
			m_genre[0] = 255;

			if (!strcmp((const char*)"TAG", (const char*)acBuf)) {
				m_nID3IsValid = 1;
				m_DataSize -= 128;
				fseek(m_pFile, -125, SEEK_END);
				fread(m_strCTitle, 1, 30, m_pFile);
				m_strCTitle[30] = '\0';
				fread(m_strCArtist, 1, 30, m_pFile);
				m_strCArtist[30] = '\0';
				fread(m_strCAlbum, 1, 30, m_pFile);
				m_strCAlbum[30] = '\0';
				fread(m_strYear, 1, 4, m_pFile);
				m_strYear[4] = '\0';
				fread(m_strComment, 1, 30, m_pFile);
				m_strComment[30] = '\0';
				if (m_strComment[28] == '\0') {
					m_track[0] = m_strComment[29];
				}
				fread(m_genre, 1, 1, m_pFile);
				
				unpad(m_strCTitle);
				unpad(m_strCArtist);
				unpad(m_strCAlbum);
				unpad(m_strYear);
				unpad(m_strComment);

				m_strTitle = m_strCTitle;
				m_strArtist = m_strCArtist;
				m_strAlbum = m_strCAlbum;
			}
		}
	}
	
	return bResult;
}

//------------------------------------------------------------------------------

char*
MP3Info::unpad(char* str)
{
	char *pos = str + strlen(str) - 1;
	while(isspace(pos[0])) {
		(pos--)[0] = 0;
	}
	
	return str;
}

//------------------------------------------------------------------------------

char*
MP3Info::pad(char* str, int length)
{
	int l;

	l = strlen(str);
	while (l < length) {
		str[l] = ' ';
		l++;
	}

	str[l] = '\0';
	return str;
}

//------------------------------------------------------------------------------

bool
MP3Info::GetFirstHeader(long nStartPos)
{
	long valid_start = 0;

	MP3Header h, h2;
	
	fseek(m_pFile, nStartPos, SEEK_SET);
	while (1) {
		int c;

		while ((c = fgetc(m_pFile)) != 255 && (c != EOF));

		if (c == 255) {
			ungetc(c, m_pFile);
			valid_start = ftell(m_pFile);

			int l;
			if ((l = GetHeader(&h))) {
				fseek(m_pFile, l - FRAME_HEADER_SIZE, SEEK_CUR);
				int k;
				for (k = 1; (k < MIN_CONSEC_GOOD_FRAMES) &&
						 (m_DataSize - ftell(m_pFile) >= FRAME_HEADER_SIZE);
					 k++) {

					if (!(l = GetHeader(&h2))) break;
					if (!SameConstant(&h, &h2)) break;
					fseek(m_pFile, l - FRAME_HEADER_SIZE, SEEK_CUR);
				}

				if (k == MIN_CONSEC_GOOD_FRAMES) {
					fseek(m_pFile, valid_start, SEEK_SET);
					memcpy(&m_Header, &h2, sizeof(MP3Header));
					m_nHeaderIsValid = 1;
					return true;
				}
			}
		}
		else {
			return false;
		}
	}
	
	return false;
}

//------------------------------------------------------------------------------

int
MP3Info::GetHeader(MP3Header* pH)
{
	unsigned char buffer[FRAME_HEADER_SIZE];

	if (fread(&buffer, FRAME_HEADER_SIZE, 1, m_pFile) < 1) {
		pH->m_nSync = 0;
		return 0;
	}

	pH->m_nSync = (((int)buffer[0] << 4) | ((int)(buffer[1] & 0xE0) >> 4));
	
	if (buffer[1] & 0x10) {
		pH->m_nVersion = (buffer[1] >> 3) & 1;
	}
	else {
		pH->m_nVersion = 2;
	}

	pH->m_nLayer = (buffer[1] >> 1) & 3;

	if ((pH->m_nSync != 0xFFE) || (pH->m_nLayer != 1)) {
		pH->m_nSync = 0;
		return 0;
	}

	pH->m_nCRC = buffer[1] & 1;
	pH->m_nBitRate = (buffer[2] >> 4) & 0x0F;
	pH->m_nFreq = (buffer[2] >> 2) & 0x3;
	pH->m_nPadding = (buffer[2] >> 1) & 0x1;
	pH->m_nExtension = (buffer[2]) & 0x1;
	pH->m_nMode = (buffer[3] >> 6) & 0x3;
	pH->m_nModeExtension = (buffer[3] >> 4) & 0x3;
	pH->m_nCopyright = (buffer[3] >> 3) & 0x1;
	pH->m_nOriginal = (buffer[3] >> 2) & 0x1;
	pH->m_nEmphasis = (buffer[3]) & 0x3;

	int nResult = FrameLength(pH);
	if (nResult < MIN_FRAME_SIZE) {
		nResult = 0;
	}
	
	return nResult;
}

//------------------------------------------------------------------------------

int
MP3Info::FrameLength(MP3Header* pH)
{
	int nResult = 1;

	if (pH->m_nSync == 0xFFE) {
		nResult = frame_size_index[3 - pH->m_nLayer] *
			((pH->m_nVersion & 1) + 1) *
			HeaderBitrate(pH) / HeaderFrequency(pH) +
			pH->m_nPadding;
	}
	
	return nResult;
}

//------------------------------------------------------------------------------

int
MP3Info::HeaderBitrate(MP3Header* pH)
{
	return bitrate[pH->m_nVersion & 1][3 - pH->m_nLayer][pH->m_nBitRate - 1];
}

//------------------------------------------------------------------------------

int
MP3Info::HeaderFrequency(MP3Header* pH)
{
	return frequencies[pH->m_nVersion][pH->m_nFreq];
}

//------------------------------------------------------------------------------

bool
MP3Info::SameConstant(MP3Header* pH1, MP3Header* pH2)
{
	if ((*(uint*)pH1) == (*(uint*)pH2)) {
		return true;
	}

	if ((pH1->m_nVersion == pH2->m_nVersion) &&
		(pH1->m_nLayer == pH2->m_nLayer) &&
		(pH1->m_nCRC == pH2->m_nCRC) &&
		(pH1->m_nFreq == pH2->m_nFreq) &&
		(pH1->m_nMode == pH2->m_nMode) &&
		(pH1->m_nCopyright == pH2->m_nCopyright) &&
		(pH1->m_nOriginal == pH2->m_nOriginal) &&
		(pH1->m_nEmphasis == pH2->m_nEmphasis)) {
		return true;
	}

	return false;
}

//------------------------------------------------------------------------------

int
MP3Info::HeaderBitRate(MP3Header* pH)
{
	return bitrate[pH->m_nVersion & 1][3 - pH->m_nLayer][pH->m_nBitRate - 1];
}

//------------------------------------------------------------------------------

void
MP3Info::Dump() const
{
	EM::I()->Trace() <<
		"------------------------------------------------------" << endl;
	EM::I()->Trace() << "Filename:  " << m_strFilename << endl;
	EM::I()->Trace() << "Title:  " << m_strTitle << endl;
	EM::I()->Trace() << "Artist:  " << m_strArtist << endl;
	EM::I()->Trace() << "Album:  " << m_strAlbum << endl;
	EM::I()->Trace() << "Seconds:  " << m_nSeconds << endl;
	EM::I()->Trace() <<
		"------------------------------------------------------" << endl;
}

//------------------------------------------------------------------------------

void
MP3Info::WriteTag(const Song& song)
{
	if (m_strTitle.length() > 0) {
		if (m_strTitle != song.GetTitle()) {
			EM::I()->Trace() << "mp3 title: " << m_strTitle << " song: " <<
				song.GetTitle() << endl;
			m_strTitle = song.GetTitle();
		}
	}
	else {
		m_strTitle = song.GetTitle();
	}

	if (m_strAlbum.length() > 0) {
		if (m_strAlbum != song.GetAlbum()) {
			EM::I()->Trace() << "mp3 album: " << m_strAlbum << " song: " <<
				song.GetAlbum() << endl;
			m_strAlbum = song.GetAlbum();
		}
	}
	else {
		m_strAlbum = song.GetAlbum();
	}

	if (m_strArtist.length() > 0) {
		if (m_strArtist != song.GetArtist()) {
			EM::I()->Trace() << "mp3 artist: " << m_strArtist << " song: " <<
				song.GetArtist() << endl;
			m_strArtist = song.GetArtist();
		}
	}
	else {
		m_strArtist = song.GetArtist();
	}

	// year, comment, genre, track

	strncpy(m_strCTitle, m_strTitle.c_str(), 30);
	pad(m_strCTitle, 30);
	strncpy(m_strCAlbum, m_strAlbum.c_str(), 30);
	pad(m_strCAlbum, 30);
	strncpy(m_strCArtist, m_strArtist.c_str(), 30);
	pad(m_strCArtist, 30);
	pad(m_strYear, 4);
	pad(m_strComment, 30);
	
	char buf[129];

	strcpy(buf, "TAG");
	strncat(buf, m_strCTitle, 30);
	strncat(buf, m_strCArtist, 30);
	strncat(buf, m_strCAlbum, 30);
	strncat(buf, m_strYear, 4);
	strncat(buf, m_strComment, 30);
	buf[124] = 28;
	if (m_track[0] != '\0') {
		buf[125] = '\0';
		buf[126] = m_track[0];
	}
	
	m_pFile = fopen(m_strFilename.c_str(), "rb+");
	if (m_pFile == NULL) {
		return;
	}
	
	fseek(m_pFile, -128, SEEK_END);
	fwrite(buf, 1, 128, m_pFile);
	fclose(m_pFile);
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
