// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// PlayListManager.cpp
//
// $Id: PlayListManager.cpp,v 1.16 2002/02/27 07:35:49 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "PlayListManager.h"

#include <algorithm>
#include <stdexcept>
#include <xb/AutoLock.h>
#include <xb/ErrorManager.h>
#include <xb/FileUtil.h>

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

pthread_mutex_t PlayListManager::s_ThreadLock1 = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t PlayListManager::s_ThreadLock2 = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t PlayListManager::s_ThreadLock3 = PTHREAD_MUTEX_INITIALIZER;

//------------------------------------------------------------------------------

PlayListManager::PlayListManager(const std::string& strRoot) :
	m_strRoot(strRoot)
{
	xb::FileUtil::AssertDirectoryExists(strRoot);
	
	try {
		UpdateDirectoryList();
	}
	catch (invalid_argument e) {
		EM::I()->Warn() << e.what() << endl;
		EM::I()->Trace() << "Directory " << strRoot <<
			" not found, creating it..." << endl;
		xb::FileUtil::CreateDirectory(strRoot);
		UpdateDirectoryList();
	}
	Dump();
}

//------------------------------------------------------------------------------

PlayListManager::~PlayListManager()
{

}

//------------------------------------------------------------------------------

void
PlayListManager::UpdateDirectoryList()
{
	m_vecPlayLists.clear();
	
	xb::FileUtil::GetDirectoryList(m_strRoot, m_listFiles);
	xb::FileUtil::DirectoryListConstItr itr;
	for (itr = m_listFiles.begin(); itr != m_listFiles.end(); ++itr) {
		PlayList p;
		// std::string strFilename(m_strRoot);
		// strFilename.append(*itr);
		p.SetFilename(m_strRoot + *itr);
		p.SetName(*itr);
		m_vecPlayLists.push_back(p);
	}

	sort(m_vecPlayLists.begin(), m_vecPlayLists.end(), Compare);
}

//------------------------------------------------------------------------------

void
PlayListManager::Dump() const
{
	EM::I()->Trace() << "Playlists: " << endl;
	
	int nCount = 1;
	xb::FileUtil::DirectoryListConstItr itr;
	for (itr = m_listFiles.begin(); itr != m_listFiles.end(); ++itr) {
		EM::I()->Trace() << nCount++ << ":  " << *itr << endl;
	}
}

//------------------------------------------------------------------------------

PlayList*
PlayListManager::GetPlayListFromFile(const std::string& strFilename)
{
	PlayListVec::iterator itr;

	for (itr = m_vecPlayLists.begin(); itr != m_vecPlayLists.end(); ++itr) {
		if (itr->GetFilename() == strFilename) {
			EM::I()->Noise() << "found playlist." << endl;
			return itr;
		}
	}

	EM::I()->Warn() << "no playlist found" << endl;
	return NULL;
}

//------------------------------------------------------------------------------

PlayList*
PlayListManager::GetPlayList(const std::string& strName)
{
	xb::Filename strFilename = GetFilename(strName);
	EM::I()->Noise() << "playlist filename:  " << strFilename << endl;
	
	return GetPlayListFromFile(strFilename);
}

//------------------------------------------------------------------------------

xb::Filename
PlayListManager::GetFilename(const std::string& strName) const
{
	return m_strRoot + xb::FileUtil::CleanFilename(strName);
}

//------------------------------------------------------------------------------

void
PlayListManager::Rename(const std::string& strOriginal,
						const std::string& strNew)
{
	PlayList* pList = GetPlayList(strOriginal);
	if (pList != NULL) {
		pList->Rename(GetFilename(strNew));
		UpdateDirectoryList();
	}
	else {
		EM::I()->Error() << "couldn't find playlist " << strOriginal << endl;
		throw(xb::file_error("couldn't find playlist " + strOriginal));
	}
}

//------------------------------------------------------------------------------

void
PlayListManager::DeletePlayList(const std::string& strName)
{
	PlayList* pList = GetPlayList(strName);
	if (pList != NULL) {
		pList->Delete();
	}
	else {
		EM::I()->Error() << "couldn't find playlist " << strName << endl;
		throw(xb::file_error("couldn't find playlist " + strName));
	}
}

//------------------------------------------------------------------------------

PlayList*
PlayListManager::NewPlayList()
{
	xb::AutoLock lock(&s_ThreadLock1, "NewPlayList()");

	return NewPlayList("New Playlist");
}

//------------------------------------------------------------------------------

PlayList*
PlayListManager::NewPlayList(const std::string& strName)
{
	xb::AutoLock lock(&s_ThreadLock2, "NewPlayList(name)");

	std::string strBase = GetFilename(strName);
	std::string strFilename = xb::FileUtil::GetAvailableFile(strBase);
	EM::I()->Noise() << "new playlist filename:  " << strFilename << endl;
	xb::FileUtil::CreateFile(strFilename);
	UpdateDirectoryList();

	return GetPlayListFromFile(strFilename);
}

//------------------------------------------------------------------------------

PlayList*
PlayListManager::NewPlayList(const std::string& strName,
							 const SongList& listSongs)
{
	xb::AutoLock lock(&s_ThreadLock3, "NewPlayLisT(name, list)");

	PlayList* pList = NewPlayList(strName);
	AssertNotNull(pList);
	
	for (SongList::const_iterator itr = listSongs.begin();
		 itr != listSongs.end(); ++itr) {

		pList->Add(*itr);
	}

	pList->Save();

	return pList;
}

//------------------------------------------------------------------------------

bool
PlayListManager::Compare(const PlayList& p1, const PlayList& p2)
{
	return p1.GetName() < p2.GetName();
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
