// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// Song.cpp
//
// $Id: Song.cpp,v 1.18 2002/02/26 19:30:27 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "Song.h"

#include <xb/ErrorManager.h>

#include "Controller.h"
#include "Info.h"
#include "System.h"
#include "Util.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

Song::Song()
{
	Song("", "", "", "", "", 0, false, false);
}

//------------------------------------------------------------------------------

Song::Song(const std::string& strArtist,
		   const std::string& strAlbum,
		   const std::string& strTitle,
		   const std::string& strTrack,
		   const std::string& strTime,
		   unsigned int nSeconds,
		   bool bSelected,
		   bool bPlaying) :
	m_nID(0),
	m_strArtist(strArtist),
	m_strAlbum(strAlbum),
	m_strTitle(strTitle),
	m_strTrack(strTrack),
	m_strTime(strTime),
	m_nSeconds(nSeconds),
	m_bSelected(bSelected),
	m_bPlaying(bPlaying)
{
	
}

//------------------------------------------------------------------------------

Song::Song(const xb::Filename& strFilename)
{
	m_strFilename = strFilename;

	Info* pInfo = System::GetController()->GetInfoPluginForFile(m_strFilename);
	if (pInfo == NULL) {
		EM::I()->Warn() << "no plugin for " << m_strFilename << endl;
		return;
	}
	
	SetArtist(pInfo->GetArtist());
	SetAlbum(pInfo->GetAlbum());
	SetTitle(pInfo->GetTitle());
	SetTrack(pInfo->GetTrack());
	SetSeconds(pInfo->GetSeconds());
	SetTime(Util::GetTimeStr(pInfo->GetSeconds()));
	SetSelected(false);
	SetPlaying(false);
}

//------------------------------------------------------------------------------

Song::~Song()
{

}

//------------------------------------------------------------------------------

bool
Song::SatisfiesSearch(const std::string& strSearch) const
{
	bool bResult = true;
	
	if (strSearch.length() > 0) {
		if (m_strArtist.find(strSearch) == std::string::npos &&
			m_strTitle.find(strSearch) == std::string::npos &&
			m_strAlbum.find(strSearch) == std::string::npos) {
			
			bResult = false;
		}
	}

	return bResult;
}

//------------------------------------------------------------------------------

void
Song::Dump() const
{
	EM::I()->Trace() << "ID: " << m_nID << endl;
	EM::I()->Trace() << "Artist: " << m_strArtist << endl;
	EM::I()->Trace() <<"Album: " << m_strAlbum << endl;
	EM::I()->Trace() << "Title: " << m_strTitle << endl;
	EM::I()->Trace() << "Track: " << m_strTrack << endl;
	EM::I()->Trace() <<"Time: " << m_strTime << endl;
	EM::I()->Trace() << "Filename: " << m_strFilename << endl;
	EM::I()->Trace() << "Seconds: " << m_nSeconds << endl;
	EM::I()->Trace() << "Selected: " << m_bSelected << endl;
	EM::I()->Trace() << "Playing: " << m_bPlaying << endl;
}

//------------------------------------------------------------------------------

std::ifstream&
operator>>(std::ifstream& is, Song& song)
{
	// first character is delimiter
	char cDelim = '*';
	// is.get(cDelim);

	EM::I()->Noise() << "delim: " << cDelim << endl;

	// PATRICK: what is a safe value for nsize?
	int nSize = 255;
	char s[nSize];

	is.getline(s, nSize, cDelim);
	song.m_nID = atoi(s);
	is.getline(s, nSize, cDelim);
	song.m_strArtist = s;
	is.getline(s, nSize, cDelim);
	song.m_strAlbum = s;
	is.getline(s, nSize, cDelim);
	song.m_strTitle = s;
	is.getline(s, nSize, cDelim);
	song.m_strTrack = s;
	is.getline(s, nSize, cDelim);
	song.m_strTime = s;
	is.getline(s, nSize, cDelim);
	song.m_strFilename = std::string(s);
	is.getline(s, nSize, cDelim);
	song.m_nSeconds = atoi(s);

	// go to the end of the line now...
	// is.getline(s, nSize);
	
	return is;
}

//------------------------------------------------------------------------------

std::ofstream&
operator<<(std::ofstream& os, const Song& song)
{
	// PATRICK: warning, if any of the data contains the delimiter, this
	// is completely wrong...need to come up with something better...

	char cDelim = '*';

	os << song.m_nID;
	
	std::string s;
	s = song.m_strArtist.length() ? song.m_strArtist : " ";
	// os << cDelim << s;
	os << cDelim << s;

	s = song.m_strAlbum.length() ? song.m_strAlbum : " ";
	os << cDelim << s;

	s = song.m_strTitle.length() ? song.m_strTitle : " ";
	os << cDelim << s;

	s = song.m_strTrack.length() ? song.m_strTrack : " ";
	os << cDelim << s;

	s = song.m_strTime.length() ? song.m_strTime : " ";
	os << cDelim << s;

	s = song.m_strFilename.length() ? song.m_strFilename : std::string(" ");
	os << cDelim << s;

	int n = song.m_nSeconds > 0 ? song.m_nSeconds : 0;
	os << cDelim << n;
	
	// os << cDelim << endl;
	os << cDelim;
	
	return os;
}

//------------------------------------------------------------------------------

bool
Song::operator==(const Song& song) const
{
	bool bResult = false;
	if (m_strFilename == song.m_strFilename) {
		bResult = true;
	}

	return bResult;
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------

